r"""We describe the class ``MM_Basics`` and its subclass ``MM_Const``
 

Class ``MM_Basics`` in module ``mmgroup.dev.mm_basics.mm_basics`` defines 
several constants that describe the organization of vectors of integers
modulo :math:`p` used in the representation modulo :math:`\rho_p`. Here
several integers modulo  :math:`p` are stored in a single unsigned
integer of type ``uint_mmv_t``. Type  ``uint_mmv_t`` is equal to the
standard C integer type ``uint64_t`` on a 64-bit system. (For a 32-bit
system that type may be changed to ``uint32_t`` by adjusting the
variable ``INT_BITS`` in file ``config.py`` of the root directory).


    .. warning::
       The 32-bit version of the project (i.e. setting ``INT_BITS = 32``) 
       is deprecated and probably no longer functional!

       If the C compiler used on a 32-bit platform supports
       integers of type ``uint64_t`` then the project may be
       compiled with setting ``INT_BITS = 64``.

    .. table:: Constants used for generating representations of the monster
      :widths: 25 75

      ==================== =====================================================
      Name                 Value
      ==================== =====================================================
      ``FIELD_BITS``       Number of bits used to store an integer modulo ``P``.
                           This is the smallest power of two greater than or
                           equal to ``P_BITS``.
      ``INT_BITS``         Number of bits available in an unsigned integer of
                           type ``uint_mmv_t``. Depending on the value
                           ``INT_BITS`` in file ``config.py`` this is 32 or 64.
      ``INT_FIELDS``       Number of integers modulo ``P`` that is stored in a
                           variable of type ``uint_mmv_t``. ``INT_FIELDS`` is 
                           a power of two and equal to ``INT_BITS/FIELD_BITS``.
      ``LOG_FIELD_BITS``   Binary logarithm of the value ``FIELD_BITS``.
      ``LOG_INT_BITS``     Binary logarithm of the value ``INT_BITS``.
      ``LOG_INT_FIELDS``   Binary logarithm of the value ``INT_FIELDS``.
      ``LOG_V24_INTS``     Binary logarithm of the value ``V24_INTS``.
      ``LOG_V64_INTS``     Binary logarithm of the value ``V64_INTS``.
      ``MVV_ENTRIES``      Number of integers modulo ``P`` contained in a
                           vector of the representation of the monster,
                           including unused entries. 
      ``MVV_INTS``         Number of integers of type ``uint_mmv_t`` required
                           to store a vector of the representation of the
                           monster. This value depends on ``P``.
      ``P``                The modulus of the current representation being
                           generated. ``P`` is equal to ``2**P_BITS - 1``.
      ``P_BITS``           Bit length of modulus ``P``.
      ``V24_INTS``         Number of integers of type ``uint_mmv_t`` used up
                           to store a vector of ``24`` integers modulo ``P``.
                           ``V24_INTS`` is always a power of two.
      ``V64_INTS``         Number of integers of type ``uint_mmv_t`` used
                           to store a vector of ``64`` integers modulo ``P``.
                           ``V64_INTS`` is always a power of two.
      ``V64_INTS_USED``    Number of integers of type ``uint_mmv_t`` actually
                           used to store a vector of ``24`` integers modulo 
                           ``P``. This may be less than ``V24_INTS``.
      ==================== =====================================================

Class ``MM_Const`` in module ``mmgroup.dev.mm_basics.mm_basics`` is
a table-providing class for the C functions used by python extension
``mmgroup.mm``. Most C functions in that module take the modulus ``p`` as 
a parameter. They need fast access to the constants given in the table
above for all legal values of ``p``. Using the code generator with the 
tables and directives provided by class ``MM_Const``, we can easily 
obtain these values for any legal ``p`` as in the following example:

.. code-block:: c

   // Return a nonzero value if p is a bad modulus,  
   // i.e. not p = 2**k - 1 for some 2 <= k <= 8
   #define mm_aux_bad_p(p) (((p) & ((p)+1)) | (((p)-3) & ((0UL-256UL))))

   // Create a precomputed table containing the constants for modulus p
   // %%USE_TABLE
   static const uint32_t MMV_CONST_TABLE[] = {
   // %%TABLE MMV_CONST_TAB, uint32
   };

   int do_something_for_modulus_p(uint 32_t p, ...)
   {
       uint32_t modulus_data, p_bits, field_bits;
       // Reject any illegal modulus p
       if (mm_aux_bad_p(p)) return -1;
       // Load the constants for modulus p to variable modulus_data
       // %%MMV_LOAD_CONST  p, modulus_data;
       // Load value P_BITS for modulus P to p_bits (using modulus_data)
       p_bits = %{MMV_CONST:P_BITS,modulus_data};
       // Load value FIELD_BITS for modulus P to field_bits 
       field_bits = %{MMV_CONST:FIELD_BITS,modulus_data};
       // Now do the actual work of the function using these values
       ...
   }
 

"""
from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals

import sys
import os
import collections
import warnings
from numbers import Integral
import re
import ast
import functools
from functools import reduce, partial
from operator import __or__



from mmgroup.bitfunctions import bitlen, hibit, lmap
from mmgroup.dev import config
from mmgroup.generate_c import format_item, c_snippet
from mmgroup.generate_c import UserDirective, UserFormat

INT_BITS = config.INT_BITS


########################################################################
# Class MM_Basics and auxiliary functions for that class
########################################################################

UINT_SUFFIX = "UL" if INT_BITS == 32 else "ULL"

def c_hex(x):
    """Convert the integer x to a useful unsigned C hex literal

    It is assumed that no integer type has more bits that
    an integer of type uint_mmv_t.
    """
    #print("c_hex", x, type(x))
    h = hex(x & ((1 << INT_BITS) - 1))
    while h[-1:] == "L": h = h[:-1] # for python 2
    return h + UINT_SUFFIX

def smask(value, fields, width):
    """Create a simple bit mask constant

    For 0 <= i and 0 <= j < 'width', the bit of that constant at 
    position i * 'width' + j is set if both, bit i of integer 'fields'
    and bit j of integer 'value' are set. A bit at 
    position >= INT_BITS is never set.
    
    Any of the arguments 'value' and 'fields' may either be an 
    integer or anything iterable that yields a list of integers. 
    Then this list is intepreted as a list of bit positions. A bit 
    of the argument is set if its position occurs in that list and 
    cleared otherwise.
    """
    try:
        value = reduce(__or__, [1 << i for i in value], 0)
    except TypeError:
        pass 
    try:
        fields = reduce(__or__, [1 << i for i in fields], 0)
    except TypeError:
        pass
    assert width > 0 
    value &= (1 << width) - 1
    result = sum(value << (width * i)
        for i in range(INT_BITS//width) if (fields >> i) & 1)
    return result & ((1 << INT_BITS) - 1)

def smask_default(default_width, value, fields = -1, width = None):
    if width is None:
        width = default_width
    return smask(value, fields, width)


def shl_fix(data, i):
    """Generate expression <data> << i for a fixed signed int i

    Generate <data> >> (-i)  in case i < 0.
    """ 
    data = str(data)
    try:
       d = int(data)
       return c_hex(d >> -i if i < 0 else d << i) 
    except:
        if i == 0:
            return "({0})".format(data)
        if i > 0:
            return "(({0}) << {1})".format(data, i)
        if i < 0:
            return "(({0}) >> {1})".format(data, -i)
          



class MM_Basics(object):
    """The basic table-providing class for vectors in :math:`\rho_p` 

    This class provides the constants given in the table above.
    It is an abstract class. Subclasses of this class  may provide
    tables and directives for dealing with vectors in :math:`\rho_p`.
    """
    MMV_ENTRIES = 759 * 64 + (3 * 24 + 3 * 2048) * 32
    INT_BITS = INT_BITS
    UINT_SUFFIX = "UL" if INT_BITS == 32 else "ULL"
    # must be 32 or 64 for uint32_t or uint64_t
    assert INT_BITS in (32, 64)
    LOG_INT_BITS = hibit(INT_BITS)
    tables = { 
        "GENERATE_CODE" : True,
        "INT_BITS": INT_BITS,
        "LOG_INT_BITS": LOG_INT_BITS,
        "MMV_ENTRIES" : MMV_ENTRIES,
        "hex": UserFormat(hex, arg_fmt = c_hex),
        "shl": UserFormat(shl_fix, "si"),
    }
    table_dict = {}
    _sizes = {}
    const_format = "uint"+str(INT_BITS) 
    p = 0  # p = 0 means that no modulus p has been specified

    @classmethod
    def hex(cls, x):
        """Convert the integer x to a useful unsigned C hex literal

        It is assumed that no integer type has more bits that
        an integer of type uint_mmv_t.
        """
        return c_hex(x)

    @classmethod
    def sizes(cls, p):
        """Return dictionary of constants for a specific p.""" 
        try:
            return cls.table_dict[p]
        except:
            assert p > 1 and p & (p + 1) == 0, str(p)
            d = cls.tables.copy()
            d["P"] = p
            d["P_BITS"] = P_BITS = bitlen(p)
            FIELD_BITS = P_BITS
            while (FIELD_BITS & (FIELD_BITS - 1)): 
                FIELD_BITS += 1 
            d["FIELD_BITS"] = FIELD_BITS
            d["LOG_FIELD_BITS"] = hibit(FIELD_BITS)
            d["INT_FIELDS"] = INT_FIELDS = cls.INT_BITS // FIELD_BITS
            d["LOG_INT_FIELDS"] = hibit(INT_FIELDS)
            V24_INTS = 32 // INT_FIELDS
            d["V24_INTS"] = V24_INTS
            d["LOG_V24_INTS"] = hibit(V24_INTS)
            d["V24_INTS_USED"] = V24_INTS - (V24_INTS >> 2)
            V64_INTS = 64  // INT_FIELDS
            d["V64_INTS"] = V64_INTS
            d["LOG_V64_INTS"] = hibit(V64_INTS)
            d["MMV_INTS"] = 3 * (2048 + 24) * V24_INTS + 759 * V64_INTS
            partial_smask = partial(smask_default, FIELD_BITS)
            d["smask"] = UserFormat(partial_smask, fmt = c_hex)
            cls.table_dict[p] = d
            return d



    def snippet(self, source, *args, **kwds):
        """Generate a C code snippet

        This method calls function ``c_snippet`` in module
        ``mmgroup.generate_c`` to generate a C code snippet.
        Parameters are as in function ``c_snippet``. The method
        returns the generated C code snippet as a string.

        In addition, all tables and directives available in the given 
        instance of the table-providing class are also passed as
        arguments to function ``c_snippet``.
        """
        tables = {"directives" : self.directives, **self.tables, **kwds}
        return c_snippet(source, *args, **tables)





########################################################################
# Class MM_Const 
########################################################################


# Auxilary function for defining methods like MM_Const().P_BITS(p) etc.
_attr_from_table = lambda name, p : MM_Basics.sizes(p)[name]


class MM_Const(MM_Basics):
    """This is the basic table-providing class for module ``mmgroup.mm``

    The main purpose of this class is to provide the constants 
    defined in class ``MM_Basics, for a variable modulus ``p`` as
    shown in the example above.

    This class provides the directive ``MMV_CONST_TAB`` for generating
    all tables, and the directive ``MMV_LOAD_CONST`` for storing the 
    table of constants, for a specific modulus ``p``, in an integer 
    variable. The string formatting function ``MMV_CONST`` can be
    used for extracting a specific constant from that variable, as 
    indicated in the example above.

    Internally, we use a deBruijn sequence to translate the value ``p``
    to an index for the table generated via the directive 
    ``MMV_CONST_TAB``. 

    Constants not depending on the modulus ``p``, such as ``INT_BITS``, 
    ``LOG_INT_BITS``, and ``MMV_ENTRIES`` are available as attributes 
    of class ``MM_Const``. They can also be coded with the code 
    generator directly via string formatting, e.g.::

        uint_8_t  a[%{MMV_ENTRIES}];

    For a fixed modulus ``p`` the constants depending on ``p`` can also
    be coded with the code generator via string formatting, e.g.::

        a >>= %{P_BITS:3};

    That constant also availble in the form ``MM_Const().P_BITS(3)``.


    Class ``MM_Const`` provides a string-formatting function ``shl`` 
    which generates a shift expression  Here::

         %{shl:expression, i}

    generates an expression equivalent to ``((expression) << i)``. 
    Here ``i`` must be an integer. In case ``i < 0`` we generate 
    ``((expression) >> -i)`` instead. E.g. ``%{shl:'x',-3}``
    evaluates to ``x >> 3``.

    Class ``MM_Const`` provides another string-formatting function
    ``smask`` which generates an integer constant to be used as a 
    bit mask for integers of type ``uint_mmv_t``. Here::

         %{smask:value, fields, width}

    with integers ``value``, ``fields``, and ``width`` creates such a 
    bit mask. ``For 0 <= i`` and ``0 <= j < width``, the bit of that 
    mask at  position ``i * width + j`` is set if both, bit ``i`` of 
    the integer ``fields`` and bit ``j`` of the integer ``value`` are 
    set. A bit in the mask at  position ``INT_BITS`` or higher is 
    never set.
    
    Any of the arguments ``value`` and ``fields`` may either be an 
    integer or anything iterable that yields a list of integers. 
    Then this list is interpreted as a list of bit positions. A bit 
    of that argument is set if its position occurs in that list and 
    cleared otherwise.
  
    E.g. ``%{smask:3, [1,2], 4}`` evaluates to ``0x330``.
    """
    # Names of the constants to be stored in an integer for each p
    entries = [  
        "LOG_INT_FIELDS", "INT_FIELDS", "LOG_FIELD_BITS",
        "FIELD_BITS", "P_BITS",
    ] 

    primes = [(1 << k) - 1 for k in range(2,9)]  # list of all p
    lengths = [0] * len(entries)  # list of max bit length of constants
    for p in primes:
        data =  MM_Basics.sizes(p)
        for i, name in enumerate(entries):        
             lengths[i] |= data[name]
    lengths = lmap(bitlen, lengths)
    #print ("SUM MM_Const bitlengths", sum(lengths))

    assert sum(lengths) <= 32
    pos = {}                      # pos is a dictionary of shape
    start = 0                     #  {entry_name:(start_pos,max_value)}
    for i, length in enumerate(lengths):
        pos[ entries[i] ] = (start, (1 << length) - 1)  
        start += length
    deBruijnMult = 0xe8           # deBruijn sequence
    table = [0] * 8               # This will be the table of constants
    for p in primes:              # Assemble that table
        data =  MM_Basics.sizes(p)
        index = (((p + 1) * deBruijnMult) >> 8) & 7 # scramble indices of 
                                  # table access via deBruijn sequence
        value = 0                  
        for entry, (shift, mask) in pos.items(): # assemble entry for p
            assert 0 <= data[entry] <= mask
            value += data[entry] << shift
        table[index] = value      # store entry for p in table
    #print("pos", pos)
    #rint("tbl", lmap(hex,table))
    T_NAME = "MMV_CONST_TAB"      # python name of contant table
    F_NAME = "MMV_CONST"        # function name "MMV_CONST"
    LOAD_F_NAME = "MMV_LOAD_CONST" # directive name "MMV_LOAD_CONST"

    # Some more constants are definded as  dividend/INT_FIELDS
    # with dividends for contant names given by:
    dividend = {
        "V64_INTS": 64, 
        "V24_INTS": 32,
        "MMV_INTS" : MM_Basics.MMV_ENTRIES,
    }

    

 
    def __init__(self):
        new_tables = {
           self.T_NAME : self.table,
           "P_LIST" : config.PRIMES, 
           "MMV_CONST" : UserFormat(self.f, "ss"),
           "smask":  UserFormat(smask, fmt = c_hex),
        }
        self.tables = self.tables.copy()
        self.tables.update(new_tables) 
        self.directives = {
          "MMV_LOAD_CONST" : 
              UserDirective(self.gen_get_const_table, "ss", "NAMES"),
        }
        for name in [
            "P", "P_BITS", "FIELD_BITS", "LOG_FIELD_BITS", 
            "INT_FIELDS", "LOG_INT_FIELDS", "V24_INTS", "LOG_V24_INTS",
            "V24_INTS_USED", "V64_INTS", "LOG_V64_INTS",
            ]:
            f = partial(_attr_from_table, name)
            setattr(self, name, f)
            self.tables[name] = UserFormat(f, "i")


    @classmethod
    def gen_get_const_table(cls, names, p,  const_p):
        """Code generating method for directive MMV_LOAD_CONST 

        According to rules of class TableGenerator the first argument 
        of this method may be an implicit dictionary that translates 
        the python name of the table of constants to the C name of that
        table. So names[cls.T_NAME] is the appropriate C name. 
       
        We calculate the index for the given input p and load the
        entry of that table corresponding to p (which is an integer)
        to the destination given by const_p.
        """
        s = "// Store constant table for {p} to {const_p}\n".format(
            const_p = const_p, p = p
        )    
        s += "{c} = {t}[(((({p}) + 1) * {mul}) >> 8) & 7];\n".format(
            c = const_p, p = p, t = names[cls.T_NAME],
            mul = cls.deBruijnMult
        )
        return s

    @classmethod
    def gen_get_const_expr(cls, const_name, const_p):
        """General code generating method for function MMV_CONST 

        It generates code that extracts the constant with name 
        const_name from the entry const_p for a specific p. 
        """
        sh, mask = cls.pos[const_name]
        s = "(({t}{sh}) & {mask})".format(
            t = const_p, mask = mask,
            sh = " >> " + str(sh) if sh else ""
        )
        return s

    @classmethod
    def gen_div_int_fields(cls, const_name, const_p):
        """Code generating for constants with names in self.dividend

        All these constants are defined as dividend/INT_FIELDS.
        The dividents for these contant names are taken from
        dictionary cls.dividend. The constant INT_FIELDS is calculated
        by method  gen_get_const_expr() of this class.
        """
        n_entries = cls.dividend[const_name]
        log_ = cls.gen_get_const_expr("LOG_INT_FIELDS", const_p)
        return "({0} >> {1})".format(n_entries, log_)
 
    @classmethod
    def f(cls, name, *args):
        """Code generating method for function MMV_CONST 

        It generates code that returns the constant with given
        'name'.
        """
        try:
            return cls.gen_div_int_fields(name, *args)
        except:
            return cls.gen_get_const_expr(name, *args)
      


class Mockup_MM_Const(MM_Const):
    def __init__(self):
        super(Mockup_MM_Const, self).__init__()
        old_tables = self.tables
        self.tables = {}
        self.tables.update(old_tables)
        self.tables["GENERATE_CODE"] = False







"""Test some C functions in file ``gen_leech.c``

In this script we test the C functions ``gen_leech2_start_type4``
and ``gen_leech2_start_type4`` in file  ``gen_leech.c``.

We use the terminology defined in
the document *The C interface of the mmgroup project*, 
section *Description of the mmgroup.generators extension*.

The tests in ths script are valid under the assumption that all
tests in the scripts ``test_leech2.py``, ``test_gen_type2``, and
``test_gen_type4`` marked with ``@pytest.mark.gen_xi`` have been 
passed. 
"""


from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals



from random import randint, choices #, shuffle, sample
from numbers import Integral
from collections import defaultdict

import numpy as np
import pytest

from mmgroup import XLeech2, Xsp2_Co1, PLoop, GCode, AutPL, Cocode, GcVector

from mmgroup import mat24
from mmgroup.generators import gen_leech2_type
from mmgroup.generators import gen_leech2_subtype
from mmgroup.generators import gen_leech2_op_atom
from mmgroup.generators import gen_leech2_reduce_type4
from mmgroup.generators import gen_leech2_op_word
from mmgroup.generators import gen_leech2_start_type4
from mmgroup.generators import gen_leech2_start_type24

# Standard vector in the Leech lattice mod 2 in Leech lattice encoding
# The standard fram \Omega
OMEGA = 0x800000
# The standard type-2 vector \beta
BETA = 0x200 

assert Cocode(BETA).syndrome() == GcVector(0xC)

#####################################################################
# Test function gen_leech2_start_type4
#####################################################################


#####################################################################
# Reference implementation of gen_leech2_start_type4()

def ref_leech2_start_type4(v):
    """Reference implementation for function gen_leech2_start_type4

    This is equivalent to function ``leech2_start_type4```.
    """
    v &= 0xffffff
    t = gen_leech2_subtype(v)
    if (t & 0xf0) != 0x40:
        return min(-1, -(t >> 4))
    if v == OMEGA:
        return 0
    t2 = gen_leech2_subtype(v ^ BETA)
    return t2 if (t2 & 0xf0) == 0x20 else t

#####################################################################
# Implementation of gen_leech2_start_type4() similar to C function


def suboctad_type(octad, w, coc):
    """Return suboctad type.

    Let ``octad`` be an octad, i.e. a bit vector of length 8. Let
    w = 1 (mod 2) if ``octad`` denotes an octad and ``w = 0`` 
    if ``octad`` denotes a complemented octad. Let ``coc`` be
    an even cocode vector in cocode representation.

    The function returns 0 in bit 1 of the return value if the 
    cocode word ``coc`` can be written as a subset of the octad,
    and 1 in bit 1 otherwise.

    the function returns ``1 + w + bit_weight(coc)/2`` in bit 0
    of the return value.

    Then Leech lattice vector ``x_octad + x_coc`` is of subtype 
    0x22 if the return value is zero. Otherwise it is of
    subtype 0x44 (or 0x46) if bit 1 of the return value is 
    (or 1).   
    """
    lsb = mat24.lsbit24(octad)            # ls bit of octad
    syn = mat24.cocode_syndrome(coc, lsb) # cocode syndrome
    wsub = octad & syn == syn             # wsub = 1 if coc is suboctad
    cw = mat24.cocode_weight(coc) >> 1    # cw = cocode_weight(v) / 2
    return 2 * (1 - wsub) + ((w ^ cw ^ 1) & 1)



def leech2_start_type4(v):
    """Return subtype of a Leech lattice frame ``v`` used for reduction

    The function returns the subtype of a vector ``v`` of type 4 in
    the Leech lattice modulo 2. Parameter ``v2`` must be in Leech 
    lattice encoding. The function returns the subtype of ``v`` that
    will be used for reduction in function ``gen_leech2_reduce_type4``.

    In that function we take care of the special case that ``v + v0``
    is of type 2 for a specific short vector ``v0``. 
    
    A simpler (but slower) implementation of thhis function is:

    If ``v ^ v0`` is of type 2 the return the subtype of ``v ^ v0``.
    Otherwise return the subtype of ``v``.

    The function returns 0 if ``v`` is equal to  ``Omega`` and
    a negative value if ``v`` has not type 4.

    This is a refernece implementation for function 
    ``gen_leech2_start_type4()`` in file ``gen_leech.c``.
    """
    if v & 0x7ff800 == 0:
        # Then v or v + Omega is an even cocode element.
        # Return 0 if v == Omega and -1 if v ==  0.
        if v & 0x7fffff == 0:
            return 0 if v & 0x800000 else -1 
        # Let w be the cocode weight. Return -2 if w == 2.
        if mat24.cocode_weight(v) != 4:
            return -2
        # Here v has type 4. Let v23 be the standard type-2 vector.
        # Return 0x20 if v ^ v23 has type 2 and 0x40 otherwise.
        return 0x20 if mat24.cocode_weight(v ^ 0x200) == 2 else 0x40  
    if mat24.scalar_prod(v >> 12, v):
        # Then v has type 3 and we return -3
        return -3
    if v & 0x800:
        # Then the cocode word 'coc' of v is odd.
        coc = (v ^ mat24.ploop_theta(v >> 12)) & 0xfff
        syn = mat24.cocode_syndrome(coc)
        # If 'coc' has weight 1 then v is of type 2 and we return -2.
        if (syn & (syn - 1)) == 0:
            return -2
        # Here v is of type 4. 
        # Return 0x21 if v ^ v23 is of type 2 and 0x43 otherwise.
        if (syn & 0xc) == 0xc and (v & 0x200000) == 0:
            return 0x21
        return 0x43
    # Let w be the weight of Golay code part divided by 4
    w = mat24.gcode_weight(v >> 12) 
    if w == 3:
        # Then the Golay code part of v is a docecad and we return 0x46.
        return 0x46
    # Here the Golay code part of v is a (possibly complemented) octad.
    # Add Omega to v if Golay part is a complemented octad.
    v ^= (w & 4) << 21
    # Put w = 1 if that part is an octad and w = 0 otherwise.   
    w = (w >> 1) & 1

    # Let 'octad' be the octad in the Golay code word in vector rep.
    octad = mat24.gcode_to_vect(v >> 12)
    coc = v ^ mat24.ploop_theta(v >> 12)  # cocode element of v
    # Return -2 if v is of type 2.
    sub = suboctad_type(octad, w, coc)
    if sub == 0:
        return -2
    # Return 0x22 if v ^ v23 is shsort
    if suboctad_type(octad, w, coc ^ 0x200) == 0:
        return 0x22
    # Otherwise return the subtype of v
    return 0x44 if sub & 2 else 0x42


#####################################################################
# Test data for function gen_leech2_start_type4()


def type4_testdata(ntests = 1000):
    """The function yields type-4 vectors in the Leech lattice mod 2

    These type-4 vectors are given in **Leech lattice encoding**.
    """
    testdata = [
        0, 
        0x800,
        0xC,
    ]
    for t in testdata:
         yield t
    for type_ in range(2,4):
        for i in range(5):
            yield XLeech2('r', type_).ord 
    for i in range(24):
         yield (1 << i) + ((randint(1,10) << 24))
         yield (1 << i) ^ 0xffffff
    for i in range(ntests):
        yield XLeech2('r', 4).ord
    for i in range(100):
        v =  XLeech2('r', 2) * XLeech2(BETA)
        if v.type == 4:
            return v.ord

#####################################################################
# Test function gen_leech2_start_type4()


@pytest.mark.gen_xi
def test_start_type4(ntests = 1000, verbose = False):
    """Test function gen_leech2_start_type4()

    The function obtains test vectors from generator function
    type4_testdata(); the number of test vectors given by ``ntests``.

    Each test vector v is a type-4 vector in the Leech lattice mod 2.
    The test function checks that the C function
    ``gen_leech2_start_type4(v)`` returns the same result as both
    functions, ``leech2_start_type4(v)`` and
    ``ref_leech2_start_type4(v)``.          
    """
    for n, v in enumerate(type4_testdata(ntests)):
        if verbose:
            print("Test %d, v = %s, subtype = %s, subtpe(v2) = %s" % (
                n, hex(v), hex(gen_leech2_subtype(v)),
                hex(gen_leech2_subtype(v ^ BETA)) 
        ))
        t_py = leech2_start_type4(v)
        t_ref = ref_leech2_start_type4(v) 
        assert t_py == t_ref, (hex(t_py), hex(t_ref))
        t_c = gen_leech2_start_type4(v)
        assert t_c == t_ref, (hex(t_c), hex(t_ref))
    


#####################################################################
# Test function gen_leech2_start_type24
#####################################################################


def ref_leech2_start_type24(v):
    r"""Reference implementation for function gen_leech2_start_type24

    The function returns the subtype of a vector v of type 2 in
    the Leech lattice modulo, provided that  v + \beta  is of
    type 4. It returns 0 in the special case  v = \beta + \Omega,
    and a negative value if v is not of type 2 or v + \beta
    is not of type 4.
    """
    v &= 0xffffff
    t = gen_leech2_subtype(v)
    if (t & 0xf0) != 0x20:
        return -1
    if (v & 0x7fffff) == 0x200:
        return 0 if v & 0x800000 else -1
    t2 = gen_leech2_subtype(v ^ BETA)
    return t if (t2 & 0xf0) == 0x40 else -1


def type24_testdata(ntests):
    r"""The function yields type-2 vectors in the Leech lattice mod 2

    These type-2 vectors are orthogonal to the standard type-2 vector
    \beta in the real Leech lattice and given in **Leech lattice
    encoding**.

    The function also yields some vectors in the Leech lattice mod 2
    that are not of this shape, for testing illegal input for function
    ``gen_leech2_start_type24``.
    """
    XLeech2_BETA = XLeech2(BETA)
    testdata = [
        0, 
        0x800,
        0xC,
        0x200,
        0x800200,
    ]
    for t in testdata:
         yield t
    for type_ in range(2,5):
        for i in range(5):
            yield XLeech2('r', type_).ord 
    for i in range(24):
         yield (1 << i) + ((randint(1,10) << 24))
         yield (1 << i) ^ 0xffffff
    for i in range(ntests):
        yield XLeech2('r', 2).ord
    for i in range(200):
        v =  XLeech2('r', 2)
        if (v * XLeech2_BETA).type == 4:
            return v.ord



@pytest.mark.gen_xi
def test_start_type24(ntests = 1000, verbose = 0):          
    """Test function gen_leech2_start_type24()

    The function obtains test vectors from generator function
    type24_testdata(); the number of test vectors given by ``ntests``.

    Each test vector v is a type-2 vector in the Leech lattice mod 2.
    The test function checks that the C function 
    ``gen_leech2_start_type24(v)``  returns the same value as
    function ``ref_leech2_start_type4(v)``.          
    """
    for n, v in enumerate(type24_testdata(ntests)):
        if verbose:
            print("Test %d, v = %s, subtype = %s, subtpe(v2) = %s" % (
                n, hex(v & 0xffffff), hex(gen_leech2_subtype(v)),
                hex(gen_leech2_subtype(v ^ BETA)) 
        ))
        #t_py = leech2_start_type24(v)
        t_ref = ref_leech2_start_type24(v) 
        #assert t_py == t_ref, (hex(t_py), hex(t_ref))
        t_c = gen_leech2_start_type24(v)
        assert t_c == t_ref, (hex(t_c), hex(t_ref))
    




"""Generating tables for operations on the rep of the monster

This includes also somw functions that use these table independent
of characteristic p
"""

from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals




import os
import sys
import numpy as np


from mmgroup.bitfunctions import bitweight, bitparity

from mmgroup.dev.mm_basics.mm_basics import MM_Const


try:
    # Try importing the fast C function
    from mmgroup import mat24 
except (ImportError, ModuleNotFoundError):
    # Use the slow python function if the C function is not available
    from mmgroup.dev.mat24.mat24_ref import  Mat24
    mat24 = Mat24



def make_octads_table():
    """Return numpy array A with entries of octads

    A[8*o+j] is the j-th entry of the octad with number o,
    for 0 <= o < 759. Here the entries of octad are the 
    the postions of the bits being set in that octad o,
    in natural order.
    """
    bl = [mat24.gcode_to_bit_list(mat24.octad_to_gcode(i))
                 for i in range(759)] 
    return np.array(sum(bl, []), dtype = np.uint8)


def make_perm64_table():
    """Make table for conversion from bit vector to suboctad

    Let o be an octad. Let {{i},{j}}, 0 <= i, j < 8, i != j be the 
    even cocode word which is a subset of octad  o,  such that 
    precisely the i-th and  the j-th bit  of octad  o  is set, as 
    in module mm_aux.c. Then  entry  8 * i + j  in the returned
    table is the number of the subocatad  {{i},{j}} in the standard 
    numbering given in module mm_aux.c.
    """
    step = [0, 1, 2, 4, 8, 16, 32, 63]
    return [step[i] ^ step[j] for j in range(8) for i in range(8)] 
  

def make_perm64_weights():
    """Return a 64-bit integer containing halved suboctad weights

    Bit i of the returned integer is set if the suboctade with
    number i has bit weight 2 (mod 4) and cleared if that
    suboctad has bit weight 0 (mod 4).    
    """
    return sum((((bitweight(i) + 1) >> 1) & 1) << i for i in range(64))


def make_suboctad_table():
    """Make table for conversion from suboctad to bit vector

    Let o be an octad. Let b_0,...,b_7 be the elements of octad 
    o in natural order. Let x = sum 2**i * x_i, i = 0,...,7.
    Then sum (x_i * b_i) is a suboctad as defined in function
    mat24_suboctad_to_cocode() in module mat24_functions.c.  
    In that function we also define a natural numbering of the
    suboctads of an octad. Entry x of following table is the
    number of the suboctad sum (x_i * b_i) for 0 <= x < 128.
    Note that bit b_7 can be computed from bits b_0,...,b_6
    as the even parity bit.
    """
    return [((i >> 1) ^ -bitparity(i)) & 0x3f for i in range(128)]



class MM_OctadTable:    
    octad_elements = make_octads_table()
    perm64_table = make_perm64_table()
    perm64_weights = make_perm64_weights() 
    suboctad_table = make_suboctad_table()

    tables = {
        "MM_OctadElementsTable" :  octad_elements,
        "MM_PERM64_TABLE": perm64_table,
        "MM_SUBOCTAD_TABLE": suboctad_table,
    }

    directives = {}

    
    
 





from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals


from random import shuffle, randint

from mmgroup.bitfunctions import lrange


def benes_step(n, sh, p):
    d = 1 << sh
    assert n == (n >> (sh + 1)) << (sh + 1)
    q = [None] * n
    for i in range(n):
        q[p[i]] = i                                               
    done = 0
    res0 = res1 = 0
    for i in range(n):
        j = i
        while not done & (1 << j):
            done |= 1 << j
            j = p[j]
            #next do: if (j & d): res1 |=  1 << (j & ~d)
            res1 |= ((j & d) >> sh) << (j & ~d)
            j = q[j ^ d]
            done |= 1 << j
            #next do: if (~j & d): res0 |=  1 << (j & ~d)
            res0 |= ((~j & d) >> sh) << (j & ~d)
            j = j ^ d    
    res2, res3 = res0 | (res0 << d), res1 | (res1 << d)
    for i in range(n):
        j = p[i ^ (((res2 >> i) & 1) << sh)] 
        q[i] = j ^ (((res3 >> j) & 1) << sh)
    return res0, res1, q


def benes_step_final(n, sh, p):
    d = 1 << sh
    res = 0
    assert n == (n >> (sh + 1)) << (sh + 1)
    for i in range(n):
        assert p[i] in [i, i^d]
        res |= ((i ^ p[i]) >> sh) << (i & ~d)
    return res
            


def benes_step_final3(n, sh, p):
    n3 = n//3
    assert n3 == 1 << sh
    res0 = res1 = res2 = 0
    for i in range(n3):
        t = p[i] >> sh
        t = 2 * t + ((p[i+n3] >> (sh + (t & 1))) & 1)
        t = (0x236407 >> (t << 2)) & 0xf
        res2 |=  (t & 1) << i
        res1 |=  ((t >> 1) & 1) << i
        res0 |=  ((t >> 2) & 1) << i
    return res0, res1, res2  



def check_benes_step(n, sh, p, res0, res1, q):
    d = 1 << sh
    def perm(n, res):
        p1 = [None] * n
        for i in range(n):
            if res & (1 << i):
                 p1[i], p1[i+d] = i+d, i
            elif i & d == 0:
                 p1[i], p1[i+d] = i, i+d
        return p1
    p0 = perm(n, res0)
    p1 = perm(n, res1)
    r = [None] * n
    for i in range(n):
        r[i] =  p1[p[p0[i]]]
    assert q == r, (q,r)
    for i in range(n):
        assert i & d == r[i] & d


def test_benes_step(n, sh, p):
    res0, res1, q = benes_step(n, sh, p)
    check_benes_step(n, sh, p, res0, res1, q)





def build_benes24(perm):
    """Compute modified Benes network for permutation of 24 entries

    The network consists of 9 layers. The returned array 'res' of 
    length 9 describes that network. In layer i, entry j is to be
    exchanged with entry  j + d[i], if bit j of the value res[i]
    is set. Here d[i] = 1,2,4,8,16,8,4,2,1 for i = 0,...,8.  
    """
    res = [None] * 9
    p = [None] * 24
    q = [None] * 24
    for i in range(24):
        p[i] = perm[i]       # copy permutation perm to p

    # The first and the last three layers are a standard Benes network. 
    # Do Benes netork looping algorithm steps for d = 1, 2, 4
    for sh in range(3):      
        d = 1 << sh
        for i in range(24):
            q[p[i]] = i      # q := inverse of p
        done = 0             # bit i marks that step i->p[i] is done
        res0 = 0             # initial looping transpositions
        res1 = 0             # final looping transpositions
        for i in range(24):  # Looping step for Benes network
            j = i            # j is a node not yet processed
            while not (done & (1 << j)): # while node j not done
                done |= 1 << j         # delare node j done
                j = p[j]               # j := permutation result p[j]
                # route node p[j] thru '0' part of inner Benes network
                # so we do: if (j & d): res1 |=  1 << (j & ~d)
                res1 |= ((j & d) >> sh) << (j & ~d)
                j = q[j ^ d]           # j = origin of buddy of p[j]
                done |= 1 << j         # declare that buddy done
                # route buddy thru '1' part of inner Benes network
                # so we do: if (~j & d): res0 |=  1 << (j & ~d)
                res0 |= ((~j & d) >> sh) << (j & ~d)
                j = j ^ d              # j = buddy of that origin
        res[sh] = res0       # save initial looping transposition
        res[8-sh] = res1     # save final looping transposition
        res0 |= res0 << d    # initial: exchange i with i^d if bit i set
        res1 |= res1 << d    # final: exchange i with i^d if bit i set
        for i in range(24):  # compute q = (initial) * p * (final)
            j = p[i ^ (((res0 >> i) & 1) << sh)] 
            q[i] = j ^ (((res1 >> j) & 1) << sh)
        for i in range(24):  # copy (initial) * p * (final) to p
            p[i] = q[i]
 
    # It remains to compute the 3 middle layers. They must compute
    # the permutation i -> p[i] with p[i] = i (mod 8). E.g. for i=0
    # we do the following transpositions, if (0, 8, 16) maps to:
    #
    #   ( 0,  8, 16):          (id)  *   (id)   *   (id)  // [1]
    #   ( 0, 16,  8):         (0,8)  *  (0,16)  *  (0,8)  // [0]     
    #   ( 8,  0, 16):         (0,8)  *   (id)   *   (id)  // [2]  
    #   ( 8, 16,  0):         (0,8)  *  (0,16)  *   (id)  // [3]
    #   (16,  0,  8):          (id)  *  (0,16)  *  (0,8)  // [4]  
    #   (16,  8,  0):          (id)  *  (0,16)  *   (id)  // [5]
    #          
    # For each permutation of (i, i+8, i+16) with compute a number j,
    # as indicated in square brackets above, from the bits 3 and 4 of
    # p[i] and p[i+8]. Then we use table look up for obtaining the
    # correct transpostions as in the list abvoe.   
    res0 = res1 = res2 = 0
    for i in range(8):
        j = p[i] >> 3
        j = 2 * j + ((p[i+8] >> (3 + (j & 1))) & 1)
        j = (0x236407 >> (j << 2)) & 0xf
        res2 |=  (j & 1) << i
        res1 |=  ((j >> 1) & 1) << i
        res0 |=  ((j >> 2) & 1) << i
    res[3], res[4], res[5] = res0, res1, res2
    return res










          


from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals

from functools import reduce
from operator import __xor__

import pytest

from random import randint, shuffle

from mmgroup import mat24
mat24fast = mat24


from mmgroup.bitfunctions import bit_mat_mul, v2
from mmgroup.dev.mat24.mat24theta import theta_to_basis_vector


from mmgroup.dev.mat24.mat24_ref import Mat24


#####################################################################
# Test computation of cocycle theta on basis vectors
#####################################################################

@pytest.mark.mat24
@pytest.mark.parametrize("gc", [Mat24, mat24fast])
def test_theta_grey(gc):
    """ yet to be documented!!!! """
    for v1 in range(0, 0x1000, 0x400):
        for v2 in range(0x10):
            v0 = v1 + v2
            v = gc.gcode_to_vect(v0)
            cc0 = gc.ploop_theta(v0) 
            cc = gc.cocode_to_vect(cc0)
            ccref = theta_to_basis_vector(v)
            assert cc == ccref 


@pytest.mark.mat24
@pytest.mark.parametrize("gc", [Mat24, mat24fast])
def test_theta_colored(gc):
    """ yet to be documented!!!! """
    for v0 in range(0,0x400,0x10):
        v = gc.gcode_to_vect(v0)
        cc0 = gc.ploop_theta(v0)   & 0xc0f
        cc = gc.cocode_to_vect(cc0)
        ccref = ((v | (v >> 1) | (v >> 2)) & 0x111111)
        assert cc == ccref , map(hex, [v,cc,ccref])


#####################################################################
# Test functions related to the multiplication in the Parker loop
#####################################################################


@pytest.mark.mat24
@pytest.mark.parametrize("gc, ref", [(mat24fast, Mat24)])
def test_Parker_loop(gc, ref):
    if gc is None: gc = Mat24
    for i in range(300):
        # test power map
        v1 = randint(0,0x1fff)
        c1 = gc.gcode_to_vect(v1)
        coc = gc.ploop_cocycle(v1,v1) 
        assert coc == (gc.bw24(c1) >> 2) & 1, map(hex, [i,v1, c1, coc])
        sq = gc.mul_ploop(v1,v1)
        assert sq == coc << 12, map(hex, [i, v1, c1, sq])
  
        # test commutator
        v2 = randint(0,0x1fff)
        c2 = gc.gcode_to_vect(v2)
        coc = gc.ploop_cocycle(v1,v2) 
        assert coc == gc.bw24(gc.ploop_theta(v1) & v2 & 0xfff) & 1
        comm = coc ^ gc.ploop_cocycle(v2,v1)
        assert comm == (gc.bw24(c1 & c2) >> 1) & 1, map(hex, [i,v1,c1,v2,c2,comm])
        assert comm == gc.ploop_comm(v1, v2) 
        pr = gc.mul_ploop(v1,v2)
        assert pr == v1 ^ v2 ^ (coc << 12), map(hex,[i,v1,c1,v2,c2,coc])  

        # test associator
        v3 = randint(0,0x1fff)
        c3 = gc.gcode_to_vect(v3)
        pr1 = gc.mul_ploop(pr,v3)
        pr2 = gc.mul_ploop(v1, gc.mul_ploop(v2,v3))
        computed_assoc = (pr1 ^ pr2) >> 12
        ref_assoc = gc.bw24(c1 & c2 & c3) & 1
        assert computed_assoc == ref_assoc
        other_assoc = gc.ploop_assoc(v1, v2, v3)
        assert other_assoc == ref_assoc

        if ref: 
            assert coc == ref.ploop_cocycle(v1,v2) 
            assert pr == ref.mul_ploop(v1,v2)
            assert gc.ploop_theta(v1) == ref.ploop_theta(v1)

        # test intersection
        cap = gc.ploop_cap(v1, v2)
        ref_cap = gc.vect_to_cocode(c1 & c2)
        assert cap == ref_cap

        # test exponentiation
        v4 = randint(0,0x1fff)
        v4_inv = gc.pow_ploop(v4,3)
        assert gc.mul_ploop(v4, v4_inv) == 0, map(hex, [v4, v4_inv, gc.mul_ploop(v4, v4_inv)])
        assert gc.pow_ploop(v4,4) == 0
        c4 = gc.gcode_to_vect(v4)
        assert gc.pow_ploop(v4,2) == (gc.bw24(c4) & 4) << 10
        assert gc.pow_ploop(v4,5) == v4
  
    print( "Parker Loop test passed" )



#####################################################################
# Test function mat24.ploop_solve
#####################################################################


def rand_ploop_solve_array(length):
    c = randint(0, 0xfff)
    a = [(1 << i) ^ ((c >> i) << 12) for i in range(12)]
    for i in range(150):
        j1, j2 = randint(0,11), randint(0,11)
        if j1 != j2:
            a[j1] ^= a[j2]
    shuffle(a)
    for i in range(12, length):
        c = randint(0, 0xfff) 
        v = reduce(__xor__, 
             ((a[i] & -((c>>i) & 1)) for i in range(12)))
        a.append(v)
    return a[:length]



def rand_ploop_solve_bad_array(length):
    a = [0x1000]
    while len(a) < length:
        a.append(randint(0, 0x1fff))
    c = randint(0, (1 << len(a)) - 1)
    for i in range(1, len(a)):
        a[0] ^= a[i] & -((c >> i) & 1)
    shuffle(a)
    return a



def ploop_solve_testcases():
    for length in list(range(14)) + [80, 161]:
        yield rand_ploop_solve_array(length)

def ploop_solve_bad_testcases():
    for length in list(range(3)) + [10, 11, 12, 13, 80]:
        yield rand_ploop_solve_bad_array(length)



@pytest.mark.mat24
@pytest.mark.parametrize("gc", [mat24fast])
def test_ploop_solve(gc):
    for m in ploop_solve_testcases():
        c = gc.ploop_solve(m)
        for v in m:
            assert gc.scalar_prod(c & 0xfff, v) == (v >> 12) & 1
    for m in ploop_solve_bad_testcases():
        with pytest.raises(ValueError):
            gc.ploop_solve(m)

     





"""Support for completing a permutation in the Mathieu group

A *heptad* is s subset of seven elements of the set on which that
Mathieu group acts. Under certain circumstances a mapping from one
heptad to another heptad can be completed to a unique element of
``mat_24``. The C function ``mat24_perm_from_heptads`` in file
``mat24_functions.c`` performs that task. It calls a simpler 
function ``mat24_perm_complete_heptad`` in that file which 
completes a mapping from a fixed heptad to variable heptad. 

This file contains python implementations of functions
``mat24_perm_complete_heptad`` and ``mat24_perm_from_heptads``.

The C function ``mat24_perm_to_m24num`` maps the elements of the 
Mathieu group ``Mat_24`` to the set if integers 
``0, ..., order(Mat_24)``. Function ``mat24_perm_complete_heptad`` 
is also used as a subroutine of function ``mat24_m24num_to_perm``
which compute the inverse of that mapping. This module also
contains python implementations of these functions.

Class ``HeptadCompleter`` is a table-providing class that is used
for code generation. It also exports the functionality of this
module.

Function ``hint_for_complete_heptads`` prints an explanation
of the implementation of function  ``mat24_perm_complete_heptad``,
including some precalulated data required for that implementation.

Function ``test_complete_octad`` test the correcteness of function
``mat24_perm_complete_heptad``. Here it suffices to check that
the identity permutation is generated correctly from the required
input values, and that the intersections of (possibly complemented)
octads and syndromes are indeed singletons if the ought to be
singletons.
"""


import sys
sys.path.append(r"C:\Data\projects\MonsterGit\src")

from mmgroup.dev.mat24.mat24tables import Mat24Tables as gc

#######################################################################
#######################################################################
# Finding octads suitable for function complete_heptad()
#######################################################################
#######################################################################

def lrange(*args):
    return list(range(*args))

def list_to_set24(l):
    """Convert a bit vector to an integer"""
    res = 0
    for x in l: res  ^= 1 << x
    return res & 0xffffff

def set24_to_list(v):
    """Convert an integer to a bit vector"""
    return [x for x in range(24) if v & (1 << x)]


_OCTAD_LIST = None
def octad_list():
    """Return the sorted list of all octads
    
    Each octad is given as a 24-bit integer describing a bit 
    vector in `vector`` representation. Octads are sorted 
    lexicographically, so lower bits of the integer are sorted
    with higher priority.        
    """
    global _OCTAD_LIST
    if _OCTAD_LIST is not None:
        return _OCTAD_LIST[:]
    l_oct = [set24_to_list(gc.octad_to_vect(o)) for o in range(759)]
    l_oct.sort()
    _OCTAD_LIST = [list_to_set24(l) for l in l_oct]
    return _OCTAD_LIST[:]

def superoctads(contained, not_contained, partial = [], weight = 0):
    """Return a specific list of octads
    
    The function returns a specific list of octads, where each 
    octad is given as a 24-bit integer describing a bit vector in 
    ``vector`` representation.
    
    An octad is included in the list if it satisfies the following
    requiremets.
     
      * All entries in the list ``contained ``must be contained 
        in the octad.
        
      * No entry in the list ``not_contained`` may be containe 
        in the octad.

      * Precisely ``weight`` entries of the list ``partial``
        must be in the octad.

    Each parameter ``contained `, ``not_contained``, or ``weight``
    may be a list of different integers < 24 or a 24-bit integer
    representing a bit vetor.    
    """
    if not isinstance(contained, int): 
        contained =  list_to_set24(contained)   
    if not isinstance(not_contained, int): 
        not_contained =  list_to_set24(not_contained)  
    if not isinstance(partial, int): 
        partial =  list_to_set24(partial)  
    return [x for x in octad_list() if (x & contained == contained)
        and (x & not_contained == 0)
        and (gc.bw24(x & partial) == weight) 
    ]

def l_superoctads(contained, not_contained, partial, weight):
    """Return a specific list of octads
    
    The function returns a spcific list of octads. Input parameters
    are as in function ``superoctads``, and also the same list of 
    octads is returns.
    
    But in contrast to function `superoctads`` each entry in that
    list of octad is represented as a sorted list of 8 integers
   
    THis function may be used for manually finding suitable octads 
    satisfying certain requiremnts.
    """
    l = superoctads(contained, not_contained, partial, weight)
    return [set24_to_list(v) for v in l]

def print_superoctads(contained, not_contained, partial, weight):
    """Print list of octads found by function ``superoctads``
    
    Input parameters are as in function ``superoctads``. Each octad
    is printed as a sorted list of 8 integers. Octads are printed
    innlexicographic order.
    """
    l = l_superoctads(contained, not_contained, partial, weight) 
    for x in l: print(x)
    if len(l) == 0: print("No octad found")

def hint_for_complete_heptads():
    """Print a hint for programming function ``complete_heptad``"""
    
    s = """\nHint for programming function ``complete_heptad()``    
    
Function ``complete_heptad()`` completes a partial 
permutation of 24 elements to a  permutation on 24 elements 
``0,..,24`` which is in the Mathieu group  ``Mat24``. Here the 
images of 7 elements of the set ``(0,..,24)`` must be given at 
the positions as indicated below. Then the constructed 
permutation is unique in the group `Mat24``.

Function ``complete_heptad()`` proceeds by completing pentads
(i.e. arbitrary subsets of size 5 of) to octads. If the images 
of the 5 elements of are known then the image of the octad is 
also known (up to permutations of the elements of the octad). 
By intersecting pentads and octads an their complements we 
will eventually split the set ``(0,..,24)`` into  singletons, 
so the the images of all elements of that set are known. 

We list suitable octads for performing the next intersection
step and also the singletons found after each step.
"""
    print(s)
    def add_to_found(found, new_found):
        print("\nSingletons found:", new_found, ". Useful octads:")
        return found + new_found
    found = [0,1,2,3,4,5,8]
    print("Given", found, ". Useful octads:")
    print_superoctads([9],lrange(14,24),found,5)
    found = add_to_found(found, [9])
    print_superoctads(lrange(8),[],found,6)
    print_superoctads([6],[7, 18, 19, 20, 21, 22, 23],found,5)
    found = add_to_found(found, [6,7])
    print_superoctads([10,12,14],[2,],found,5)
    found = add_to_found(found, [10,11,12,13,14])
    print_superoctads([15,16,18],[17,7],found,5)
    found = add_to_found(found, [16,17])
    print_superoctads([15,19,20],[17, 18],found,5)
    found = add_to_found(found, [15,18])
    print_superoctads([19,21,22],[15, 17, 20],found,5)
    found = add_to_found(found, [19,20])
    print_superoctads([21,23],[2,7,12,15,17,18,19,20,22],found,6)
    print("")


#######################################################################
#######################################################################
# Function complete_heptad()
#######################################################################
#######################################################################


#######################################################################
# Auxiliary functions for function complete_heptad()
#######################################################################





def lsb24(x):
    """Return binary logarithm of ``x`` if ``x`` is a power of 2"""
    res = int(x).bit_length() - 1
    assert x == 1 << res, hex(x)
    return res

def odd_syn(v):
    """Return Golay cocode syndrom of a bit vector v

    The bit vector v must have odd parity.
    """
    coc = gc.vect_to_vintern(v)
    t = gc.syndrome_table[coc & 0x7ff ]
    return ( (1 << (t & 31)) ^ (1 << ((t >> 5) & 31)) ^
             (1 << ((t >> 10) & 31)) )


#######################################################################
# Function complete_heptad()
#######################################################################


""" MOG

    0      4      8   C=12   G=16   K=20
    1      5      9   D=13   H=17   L=21
    2      6   A=10   E=14   I=18   M=22
    3      7   B=11   F=15   J=19   N=23
"""

def mat24_complete_heptad(p_io):
    """Complete a permutation p given by p_io to an element of  Mat24.
   
    p must have length 24. Entries p[i], i = 0,1,2,3,4,5,8 must make
    up a valid umbral heptad, i.e. a heptad not contained in an octad.
    p[0],...,p[5] must be contained in an octad, p[8] must not be
    contained in that octad. The other entries of input p are ignored.
   
    It can be shown that such a permutation p can be completed to
    a unique element of Mat24.
   
    If a suitable permutation p an be found, the array p_io is
    completed so that it contains permutation p, and the function
    returns 0.
   
    Otherwise the function returns a nonzero value.
    """
    ## uint_fast32_t  err, s1, s5, s015, s3, s4, s8, s01234; 
    ## uint_fast32_t  s567, s67, s9AB, s9CD, s9, s6GH, s6;
    ## uint_fast32_t  sACE, sD, sFGI, sG, sFJK, sJLM, sALN;

    err = ((p_io[0] + 8) | (p_io[1] + 8) | (p_io[2] + 8) | 
      (p_io[3] + 8) | (p_io[4] + 8) | (p_io[5] + 8) | (p_io[8] + 8));
    err &= (0 - 0x20);
    s1 = 1 << p_io[1]; 
    s5 = 1 << p_io[5];
    s015 = (1 << p_io[0]) ^ s1 ^ s5;
    s3 = 1 << p_io[3]; 
    s4 = 1 << p_io[4]; 
    s8 = 1 << p_io[8];
    s01234 = s015 ^ s5 ^ (1 << p_io[2]) ^ s3 ^ s4;
    # if err == 0 then 0 <= s01234 < 0x1000000 has odd parity
    # octad = [0, 1, 2, 3, 4, 5, 6, 7]
    s567 = odd_syn(s01234);
    err |= (s01234) & s567;
    # if err == 0 then entries [0,1,2,3,4,6] are in an octad
    err |= (s01234 | s567) & s8;
    # if err == 0 then entry 8 is not in that octad
    err |= s5 ^ (s5 & s567);
    s67 = s567 & ~s5;
    # octad = [0, 1, 2, 3, 8, 9, 10, 11]
    s9AB = odd_syn(s01234 ^ s4 ^ s8);
    # octad = [0, 1, 4, 5, 8, 9, 12, 13]
    s9CD = odd_syn(s015 ^ s4 ^ s8);
    s9 = s9AB & s9CD;
    p_io[9] = lsb24(s9);
    # octad [1, 3, 5, 6, 8, 9, 16, 17]
    s6GH = odd_syn(s1 ^ s3 ^  s5 ^ s8 ^ s9);
    s6 = s67 &  s6GH;
    p_io[6] = lsb24(s6);
    p_io[7] = lsb24(s67 & ~s6GH);
    # still needed: 
    #   err, s1, s015, s3, s8, s01234, s9AB, s9CD, s9, s6GH, s6
    # octad [0, 2, 4, 6, 8, 10, 12, 14]
    sACE = odd_syn(s01234 ^ s1 ^ s3 ^ s6 ^ s8);
    p_io[10] = lsb24(s9AB & sACE);
    p_io[11] = lsb24(s9AB & ~sACE & ~s9);
    p_io[12] = lsb24(s9CD & sACE);
    sD = s9CD & ~sACE & ~s9;
    p_io[13] = lsb24(sD);
    p_io[14] = lsb24(sACE & ~s9AB & ~s9CD);
    # still needed: 
    #   err, s1, s015, s3, s8, s6GH, s6, sACE, sD
    # octad [0, 1, 5, 6, 13, 15, 16, 18]
    sFGI = odd_syn(s015 ^ s6 ^ sD);
    sG = s6GH & sFGI;
    p_io[16] = lsb24(sG);
    p_io[17] = lsb24(s6GH & ~s6 & ~sFGI);
    # octad [0, 1, 3, 5, 8, 15, 19, 20]
    sFJK = odd_syn(s015 ^ s3 ^ s8);
    p_io[15] = lsb24(sFGI & sFJK);
    p_io[18] = lsb24(sFGI & ~sG & ~sFJK);
    # octad [0, 3, 5, 6, 16, 19, 21, 22]
    sJLM = odd_syn(s015 ^ s1 ^ s3 ^ s6 ^ sG);
    p_io[19] = lsb24(sFJK & sJLM);
    p_io[20] = lsb24(sFJK & ~sFGI & ~sJLM);
    # octad [0, 1, 5, 6, 8, 10, 21, 23]
    sALN = odd_syn(s015 ^ s6 ^ s8); 
    p_io[21] = lsb24(sALN & sJLM);
    p_io[22] = lsb24(sJLM & ~sALN & ~sFJK);
    p_io[23] = lsb24(sALN & ~sACE & ~sJLM);
    return err;

#######################################################################
#######################################################################
# Function based on function complete_heptad()
#######################################################################
#######################################################################


#######################################################################
# Compute the n-th element of the Mathieu group Mat24
#######################################################################


STD_OCTAD = gc.vect_to_octad(0xff)




def split_n_mat24(n):
    """Auxiliary function for function ``py_mat24_int_to_perm``

    This function converts the integer 0 <= n < 244823040 to
    mixed-radix notation with bases [16, 3, 20, 21, 22, 23, 24].
    It returns the list of the 8 corresponding digits, with the
    highest digit first.
    """
    n_list = []
    for d in [16, 3, 20, 21, 22, 23, 24]:
        n, r = divmod(n, d)
        n_list.append(r)
    n_list.reverse()
    return n_list

def py_mat24_int_to_perm(n):
    """Return the k-th permutation p in the Mathieu group Mat24   

    Here the elements of the Mathieu group Mat24 are numbererd
    in lexicographic order.

    Any integer 0 <= n < 244823040 is evaluated in mixed-radix with 
    bases  24, 23, 22, 21, 20, 3, 16, with valence decreasing from 
    left to right. The first five digits determine the images of
    the first five elements of the permutation. The next digit
    determines the image of entry 5. That entry has 3 possible 
    images, depending on the syndrome of the images of the first
    five elements. The final digit determines the image of element 8, 
    i.e.  the first element not in the standard octad. 

    The images of the remaining elements 6, 7 and 9,...,23  are 
    determined by calling function complete_heptad()

    This implementation of the function is easy to understand but
    not fast. Function ``mat24_int_to_perm`` is a faster version of
    this function.
    """
    assert 0 <= n < MAT24_ORDER
    # Compute the output permutation in ``p``
    p = [None] * 24
    # List of elements to be permuted by Mat24
    p_list = list(range(24))
    # Convert n to mixed-radix digit list ``n_list`` as described above.
    n_list = split_n_mat24(n)
    # Compute the images of the first five numbers 0,...,4 under p
    # from the the first five digits of the digit list.
    # After assigning an image, delete that image from ``n_list``,
    # in order to avoid repetitions.
    for i in range(5):
        index = n_list[i] 
        p[i] = p_list[index]
        del p_list[index]
    # Compute the syndrome these five images in the list ``syn``.
    # That syndrome has length 3.  
    bitmap = sum(1 << i for i in p[:5])
    cocode = gc.vect_to_cocode(bitmap)
    syn_tab = gc.syndrome_table[cocode & 0x7ff]
    syn = [(syn_tab >> i) & 31 for i in [0, 5, 10]] 
    # Select image of the number 5 from that syndrome,
    # using entry 5 f the digit list ``n_list``
    p[5] = syn[n_list[5]]
    # Let p_list be the list of number 0 <= i < 24 that are not images
    # any of the numbers of [0,1,2,3,4] and not in the syndrome.
    bitmap |= sum(1 << i for i in syn)
    p_list = [i for i in range(24) if ((1 << i) & bitmap) == 0]
    # Select image of the number 8 from that list,
    # using entry 6 f the digit list ``n_list``
    p[8] = p_list[n_list[6]]
    mat24_complete_heptad(p)
    return p


MAT24_ORDER =  244823040 
SH = 58
FACTOR = (1 << SH) // MAT24_ORDER + 1
FACTOR24 = 24 * FACTOR
DFIELD1 = 0x555555555555

def mat24_int_to_perm(n):
    """Return the k-th permutation in the Mathieu group Mat24   

    Here the elements of the Mathieu group Mat24 are numbererd
    in lexicographic order, starting with 0. 

    This is a fast version of function ``py_mat24_int_to_perm``.
    There is also a C version of this function. 
    """
    # See function ``py_mat24_int_to_perm`` for a simpler implementation
    # We document the tricks used to accelerate that implementation.
    assert 0 <= n < MAT24_ORDER
    # Compute the output permutation in ``p``
    p = [None] * 24

    # Compute ``n1`` such that ``k = n1 >> SH`` is the current digit in 
    # the mixed-radix representation of n as described in function    
    # ``py_mat24_int_to_perm``. For obtaining the next digit we subtract
    # k << SH from ``n1`` and then we  multiply ``n1`` by ``i``, 
    # for i = 23, 22, 21, 20, 3, 16 in that order. 
    n1 = FACTOR24 * n
    # Let k be the first digit of n. Store this in p[0] 
    p[0] = k = n1 >> SH  
    n1 -= k << SH      # process n1 for next digit
    # We also accumulate the images p[0],...,p[4] in a bitmap
    bitmap = 1 << k

    # For assigning the next images p[m], m = 1,2,3, we take the
    # m-th digit n_m from n and we put p[m] = n_m + d[m]. Here d[m]
    # is a difference that must be added to  n_m in order to avoid
    # repetitions in the images of p[m], 0 <= m < 5.  
    # Computation of d[m] is a bit tricky. We keep the array 
    # d[m], 0 <= m < 24 in the 64-bit integer d as a bit field, 
    # reserving two bits at positions 2*m, 2*m+1 for d[m]. After
    # selecting the image k = n_m, we put p[m] = k + d[k]. Then we
    # have to replace d[u] by d[u + 1] + 1 for u >= k. We don't 
    # change d[u] for u < j. We start with d[u] = 0 for all u. 
    # After putting p[0] = k, we put d[u] = 1 for u >= k.  
    d = DFIELD1 << (2*k)
    for i in [23, 22, 21]:
        # Obtain next digit k of n and adjust n1
        n1 = i * n1
        k = n1 >> SH
        n1 -= k << SH
        # put p[24-i] = j = k  - d[k]
        p[24-i] = j = k + ((d >> (2*k)) & 3)
        # Adjust bit field d: replace d[u] by d[u + 1] + 1 for u >= k.
        # Unfortunately, we would get an overflow in the last round. So
        # in the last round we replace replace d[u] by d[u + 1] instead,
        # and we remember the last value k in the variable ``last``. 
        mask = (1 << (2*k)) - 1
        if i > 21:
            d = (((d + DFIELD1) >> 2) & ~mask) + (d & mask) 
        else:
            d = ((d >> 2) & ~mask) + (d & mask) 
            last = k
        # Enter the image j into the bitmap
        bitmap |= 1 << j 

    # Assign the image p[4]. This works in the same way as assigning
    # the previous imgaes; but here we have to increment d[k] by one
    # in case  k >= last.
    n1 = 20 * n1
    k = n1 >> SH
    n1 -= k << SH
    p[4] = k + ((d >> (2*k)) & 3) + (k >= last)
    bitmap |= 1 << p[4] 

    # Now we have assigned the images p[m], 0 <= m < 5. An we have 
    # computed the bitmap of these images in ``bitmap``.  Next we 
    # compute the syndrome of these five images in ``syn``. Here
    # integer ``syn`` is to be interpreted as a list of three bit
    # fields of length 5. Each field contains an entry of the
    # syndrome. These entries are ordered. 
    cocode = gc.vect_to_cocode(bitmap)
    syn = gc.syndrome_table[cocode & 0x7ff]

    # Get next digit k from n. Assign the k-th entry of the
    # syndrome to p[5]
    k = (3 * n1) >> SH
    p[5] = (syn >>  (5 * k)) & 31 
   
    # Enter all entries of the syndrome into the bitmap
    bitmap |= (1 << (syn & 31)) | (1 << ((syn >> 5) & 31))
    bitmap |= (1 << ((syn >> 10) & 31))
    # Complement the bitmap. So the bitmap has 16 bits set.
    bitmap ^= 0xffffff

    # Compute ordered list of positions of the bits set in the bitmap    
    j = 0
    p1 = [None] * 24
    for i in range(24):
        p1[j] = i;              # write index i to output pos.
        j += (bitmap >> i) & 1; # increment output pos. if bitmap[i]=1

    # Let k (where k = n % 16) be the last digit of n. Assign the k-th
    # entry of that list to p[8].
    p[8] = p1[n & 15]

    # Now we have assigned p[m], m = 0,1,2,3,4,5,8; and we procceed
    # as in in function ``py_mat24_int_to_perm``.
    mat24_complete_heptad(p)
    return p








#######################################################################
# Compute the index of an element of the Mathieu group Mat24
#######################################################################




def mat24_perm_to_int(p1):
    """Convert permutation p1 in the Mathieu group Mat24 to an integer.

    This reverses member function mat24_int_to_perm(). The input 
    permutation is not checked.

    The function returns an undefined integer if permutation
    p1 is not in the Mathieu group.
    """
    # We reverse to operation of function ``mat24_int_to_perm``.
    # We compute the digits of the result n in mixed-radix
    # representation as described in function ``mat24_int_to_perm``.
    # Digits are taken from the images p1[m], m = 0,1,2,3,4,5,8 in 
    # that order.
    # Get first digit k from p1[0]
    n = k = p1[0]
    # We also accumulate the images p1[0],...,p1[4] in a bitmap
    bitmap = 1 << k

    # For obtaining the next digits n_m from  p1[m], m = 1,2,3, we
    # put  n_m = p1[m] - d[m]. Here d[m] is a difference that must 
    # be subtracted n_m in order to ensure 0 <= n_m < 24-m.
    # Computation of d[m] is a bit tricky. We keep the array 
    # d[m], 0 <= m < 24 in the 64-bit integer d as a bit field, 
    # reserving two bits at positions 2*m, 2*m+1 for d[m]. After
    # selecting the image k = p1[m], we put n_m = k - d[k]. Then we
    # have to replace d[u] by d[u + 1] + 1 for u > k. We don't 
    # change d[u] for u < k. Since a permutation has no repetitions, 
    # we may ignore d[k]. We start with d[u] = 0 for all u. 
    # After putting n_0 = p1[0], we put d[u] = 1 for u >= n_0.  
    d = DFIELD1 << (2*k)
    for i in [23, 22, 21]:
        # Store current entry of p1 in k and adjust bitmap
        k = p1[24-i]
        bitmap |= 1 << k 
        # Set digit n_k = k - d[k] and compute next step for n
        n = i * n + k - ((d >> (2*k)) & 3)
        # Adjust bit field d: replace d[u] by d[u + 1] + 1 for u >= k.
        # Unfortunately, we would get an overflow in the last round. So
        # in the last round we replace replace d[u] by d[u + 1] instead,
        # and we remember the last value k in the variable ``last``.         
        if i > 21:
            d += DFIELD1 << (2*k) 
        else:
            last = k

    # Obtain digit n_4 from p[4]. This works in the same way as
    # assigning the previous imgaes; but here we have to increment
    # d[k] by one in case  k >= last.
    k = p1[4]
    bitmap |= 1 << k 
    n = 20 * n + k - ((d >> (2*k)) & 3) - (k >= last)

    # Now we have obtained the the digis n_m, 0 <= m < 5, from the
    # images p[m], 0 <= m < 5. An we have computed the bitmap of these 
    # images in ``bitmap``.  Next we compute the syndrome of these five
    # images in ``syn``. Here integer ``syn`` is to be interpreted as a 
    # list of three bit fields of length 5. Each field contains an 
    # entry of the syndrome. These entries are ordered. 
    # Also let syn1 = syn[1] be the middle entry of that list.
    cocode = gc.vect_to_cocode(bitmap)
    syn = gc.syndrome_table[cocode & 0x7ff]
    syn1 = (syn >> 5) & 31

    # Compute next digit n_5 = k from p[5]. Put k = 2 if p1[5] > syn1,
    # k = 1 if p1[5] == syn1, and  k = 0 if p1[5] < syn1.
    k = int(p1[5] > syn1) + int(p1[5] >= syn1)
    # Enter that digit into the accumulated value n
    n = 3 * n + k

    # Enter all entries of the syndrome into the bitmap
    bitmap |= (1 << (syn & 31)) | (1 << syn1)
    bitmap |= (1 << ((syn >> 10) & 31))

    # Delete all bits at positions >= p1[8] from the bitmap. The the
    # last digit of n ist the bit weight of that bit map.
    d = gc.bw24(((1 << p1[8]) - 1) & bitmap)
    # Enter the last digit into n
    n = 16 * n + p1[8] - d
    # Change n to zero if it is too large
    n & ((n >= MAT24_ORDER) - 1); 
    return n






#######################################################################
# Compute permitatio in the Mathieu group Mat24 from heptads
#######################################################################

           
            
def mat24_perm_from_heptads(h1, h2):
    """Try to find a permutation p that maps heptad h1 to h2

    h1 and h2 must be lists of length 7 defining two umbral heptads,
    i.e. heptads not contained in an octad. If a permutation p in
    the Mathieu group Mat24 that maps h1 to h2 exists, it is unique. 

    Return permutation p if such a p exists an is unique,
    and return None otherwise.
    """
    # First find the special element of v h1 not contained in the octad
    v = 0
    for i in range(7):
        v ^= 1 << (h1[i] & 31)
    y = odd_syn(v)
    v =  lsb24(v & y)
    
    # Find position y of element v in h1
    y = 0
    for i in range(7): 
        y |= ((h1[i] != v) - 1) & i

    # Copy special element of h1 to position 8 of p1 and copy the other
    # elements of h1 to positions 0,...,6. Copy h2 similarly to p2
    p1 = h1[:7] + [None]*17
    p2 = h2[:7] + [None]*17
    p = [None] * 24
    p1[8] = p1[y]
    p1[y] = p1[6]
    p2[8] = p2[y]
    p2[y] = p2[6]

    # Complete p1 and p2 from heptad. Return error if any completion fails
    if mat24_complete_heptad(p1) | mat24_complete_heptad(p2):
        return None

    # If success, return p1**(-1) * p2
    for i in range(24):
        p[p1[i]] = p2[i]
    return p


#######################################################################
#######################################################################
# Support for the code generator
#######################################################################
#######################################################################


class HeptadCompleter(object):

    def __init__(*args):
        # Ignore arguments in constructor for compatibility
        # with older versions
        pass

    @staticmethod
    def complete_heptad(p_io):
        return  mat24_complete_heptad(p_io)

    @staticmethod
    def int_to_perm(k):
        return  mat24_int_to_perm(k)

    @staticmethod
    def perm_to_int(p):
        return  mat24_perm_to_int(p)

    @staticmethod
    def perm_from_heptads(h1, h2):
        return mat24_perm_from_heptads(h1, h2)

    @staticmethod
    def tables():
        """Return dictionary with the constant STD_OCTAD. """
        return {
            "MAT24_STD_OCTAD" : STD_OCTAD,
        }

    @staticmethod
    def directives():
        return { }


#######################################################################
#######################################################################
# Testing this module
#######################################################################
#######################################################################


def test_complete_octad():
    io = lrange(6) + [None,None,8] + [None]*15    
    complete_heptad(io)
    assert io == lrange(24)   
    print("The python function ``complete_heptad()`` is correct")

if __name__ == "__main__":
    hint_for_complete_heptads()
    test_complete_octad()
    
    


from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals



import sys
import os
import collections
import re
import warnings
from numbers import Integral
import numpy as np
from random import randint, sample


from mmgroup.structures.parse_atoms import  AtomDict      
#from mmgroup.structures.parse_atoms import eval_atom_expression        
from mmgroup.structures.parse_atoms import TaggedAtom
from mmgroup.structures.parse_atoms import ihex
from mmgroup.structures.construct_mm import iter_mm       
from mmgroup.structures.construct_mm import load_group_name     
from mmgroup.structures.construct_mm import add_to_embedded_classes     
from mmgroup.structures.abstract_mm_group import AbstractMMGroup


from mmgroup.tests.groups.auto_group import AutoGroupWord
from mmgroup.tests.groups.auto_group import AutoGroup
from mmgroup.tests.groups.auto_group import AutoGroupMulError


from mmgroup import mat24




###########################################################################
# Atoms for the group N 
###########################################################################


n_rules = 0
n_inv = 0 
n_symbols = 0




class AutPLoopAtom(TaggedAtom):
    __slots__ = 'tag', 'm24num', 'perm', 'rep', 'cocode'
    def __init__(self, tag, cocode, perm=0):
        self.tag = tag
        if isinstance(cocode, Integral):
            self.cocode = cocode & 0xfff
            if perm:
                if isinstance(perm, Integral):
                    self.m24num = perm
                else:
                    self.m24num = mat24.perm_to_m24num(perm)
                self.perm = mat24.m24num_to_perm(self.m24num)
                self.rep = mat24.perm_to_autpl(cocode, self.perm)
            else:
                self.m24num =  0
                self.perm = mat24.m24num_to_perm(0)
                self.rep = mat24.cocode_to_autpl(cocode)
        else:
            self.rep = cocode
            self.perm = mat24.autpl_to_perm(self.rep)
            self.cocode = mat24.autpl_to_cocode(self.rep) 
            self.m24num = mat24.perm_to_m24num(self.perm)

    def reduce(self, *args):
        self.cocode  &= 0xfff
        return None if self.cocode == self.m24num == 0 else self


    def as_atoms(self):
        yield 0x10000000 + self.cocode
        yield 0x20000000 + self.m24num

    def as_tuples(self):
        yield "d", self.cocode
        yield "p", self.m24num

    def str(self):
        prod = []
        c = self.cocode & 0xfff
        if c:  
            prod.append("d_%s" % ihex(c)) 
        if self.m24num: 
            prod.append("p_%d" % self.m24num)
        s = "*".join(prod)
        return s if s else "1"
    __repr__ = str

tag_dict = {
        "x": 0x30000000, 
        "y": 0x40000000, 
        "t": 0x50000000, 
        "l": 0x60000000, 
}

class xy_Atom(TaggedAtom):
    __slots__ = 'tag', 'pl'
    format_tuple = (str, ihex)

    def __init__(self, tag, pl):
        self.tag = tag
        self.pl = pl & 0x1fff

    def reduce(self, *args):
        self.pl  &= 0x1fff
        return None if self.pl == 0 else self

    def as_atoms(self):
        yield tag_dict[self.tag] + self.pl

    def as_tuples(self):
        yield self.tag, self.pl


class tl_Atom(TaggedAtom):
    __slots__ = 'tag', 'exp'
    format_tuple = (str, str)

    def __init__(self, tag, exp):
        self.tag = tag
        self.exp = exp % 3

    def reduce(self, *args):
        self.exp  %= 3
        return None if self.exp == 0 else self

    def as_atoms(self):
        yield tag_dict[self.tag] + self.exp

    def as_tuples(self):
        yield self.tag, self.exp




###########################################################################
# Generating atoms of the group N from tuples
###########################################################################

# deprecated

def gen_d(tag, c = "r"):
    cocode = c
    if isinstance(c, str):
       cocode = randint(0,0xfff)
       if "o" in c and not "e" in c:
           cocode |= 0x800 
       if "e" in c and not "o" in c:
           cocode &= ~0x800
    yield AutPLoopAtom('p', cocode & 0xfff, 0)

def gen_p(tag, perm = "r"):
    if isinstance(perm, str):
        perm = randint(0, mat24.MAT24_ORDER-1)
    else:
        if not 0 <= perm < mat24.MAT24_ORDER:
            raise ValueError("Bad permutation number for Mathieu group")
    yield AutPLoopAtom('p', 0, perm)

def gen_xy(tag, r = "r"):
    pl = randint(0, 0x1fff) if isinstance(r, str) else r & 0x1fff
    yield xy_Atom(tag, pl & 0x1fff)


def gen_z(tag, r = "r"):
    pl = randint(0, 0x1fff) if isinstance(r, str) else r & 0x1fff
    pl = mat24.pow_ploop(pl, 3)
    yield xy_Atom('y', pl)
    yield xy_Atom('x', pl)



def gen_tl(tag, e = "r"):
    if isinstance(e, str):
        e = randint( (0 if "z" in e else 1), 2) 
    else:
        e = e % 3
    yield tl_Atom(tag, e) 




###########################################################################
# Rules deleting neutral elements
###########################################################################


def rule_x0(group, word):
    x = word[0].pl & 0x1fff
    if x == 0:
        return []
    else:
        raise AutoGroupMulError

def rule_p0(group, word):
    d, p = word[0].cocode & 0xfff, word[0].m24num
    if (d | p)  == 0:
        return []
    else:
        raise AutoGroupMulError


def rule_tl0(group, word):
    exp = word[0].exp & 3
    if exp == 0:
        return []
    else:
        raise AutoGroupMulError


###########################################################################
# Rules for the group N 
###########################################################################



def rule_pp(group, word):
    global n_rules
    n_rules += 1
    rep = mat24.mul_autpl(word[0].rep, word[1].rep)
    return  [AutPLoopAtom('p', rep)]


def rule_yp(group, word):
    global n_rules
    n_rules += 1
    y, p = word[0], word[1]
    pl = mat24.op_ploop_autpl(y.pl, p.rep)
    if p.cocode & 0x800:
        pl = mat24.pow_ploop(pl, 3)
        return [p, xy_Atom('y', pl), xy_Atom('x', pl)]
    else:
        return [p, xy_Atom('y', pl)]


def rule_xp(group, word):
    global n_rules
    n_rules += 1
    x, p = word[0], word[1]
    pl = mat24.op_ploop_autpl(x.pl, p.rep)
    return [p, xy_Atom('x', pl)] 





 
# Extra relations given by Kernels K0 and K1: 
# K0:  y_(-1) = x_(-Omega), y_Omega = x_(-1), y_(-Omega) = x_Omega
# K1:  y_(-1) = x_(-Omega)
# Anything apart from kernel K0 isno longer supported!!!
dict_kernel0_yx = {0:0, 0x1000:0x1800, 0x800:0x1000, 0x1800:0x800, }
#dict_kernel1_yx = {0:0, 0x1000:0x1800}

dict_kernel0_xy = {0:0, 0x1000:0x800, 0x800:0x1800, 0x1800:0x1000, }
#dict_kernel1_xy = {0:0, 0x1000:0x800}


#kernel_data = [
#   (0x1800, dict_kernel0_yx, dict_kernel0_xy),  # for kernel K_0
#   (0x1000, dict_kernel1_yx, dict_kernel1_xy),  # for kernel K_1
#]



def kernel_rule_yx(group, word):
    global n_rules
    n_rules += 1
    y, x = word[0].pl & 0x1fff,  word[1].pl & 0x1fff
    mask, table_yx, table_xy = 0x1800, dict_kernel0_yx, dict_kernel0_xy
    if y & ~mask and not x & ~mask:
        y ^= table_xy[x]
        return  [xy_Atom('y', y)]
    else:
        x ^= table_yx[y & mask]
        y_new = y & ~mask 
        if x == 0:
            return  [xy_Atom('y', y_new)]
        elif y_new == y:
            raise AutoGroupMulError
        else:
            return [ xy_Atom('y', y_new),  xy_Atom('x', x) ]


def kernel_rule_y(group, word):
    global n_rules
    n_rules += 1
    y = word[0].pl & 0x1fff
    mask, table  =  0x1800, dict_kernel0_yx
    if y & mask == y:
        return [ xy_Atom('x', table[y]) ]
    else:
        raise AutoGroupMulError

    mask = y & mask
    x = table[y & mask]
    y_new = y & ~mask & 0x1fff
    #print("kernel", hex(y), hex(y_new))
    if y == 0:
        return []
    if y_new == y:
        raise AutoGroupMulError
    return [ xy_Atom('y', y_new),  xy_Atom('x', x) ]


    
def rule_xx(group, word):
    global n_rules
    n_rules += 1
    x1, x2 = word[0], word[1]
    delta = AutPLoopAtom('p', mat24.ploop_cap(x1.pl, x2.pl), 0)
    x = xy_Atom(x1.tag, mat24.mul_ploop(x1.pl, x2.pl))
    return [delta, x]

rule_yy = rule_xx


def rule_xy(group, word):
    global n_rules
    n_rules += 1
    x, y = word[0], word[1]
    delta = AutPLoopAtom('p', mat24.ploop_cap(x.pl, y.pl), 0)
    comm =  mat24.ploop_comm(x.pl, y.pl) << 12
    x1 = xy_Atom(x.tag, x.pl ^ comm)
    y1 = xy_Atom(y.tag, y.pl ^ comm)
    return [delta, y1, x1]



def rule_tt(group, word):
    global n_rules
    n_rules += 1
    exp = (word[0].exp + word[1].exp) % 3
    return  [tl_Atom(word[0].tag, exp)] if exp else []


def rule_pt(group, word):
    global n_rules
    n_rules += 1
    p, t = word[0], word[1]
    exp = 3 - t.exp  if p.cocode & 0x800 else t.exp
    t1 = tl_Atom('t', exp % 3)
    return [t1, p]

def rule_xt(group, word):
    global n_rules
    n_rules += 1
    x, t = word[0], word[1]
    assert t.exp in [1,2]
    if t.exp == 2:
        pl = mat24.pow_ploop(x.pl, 3)
        x, y = xy_Atom('x', pl), xy_Atom('y', pl)
        return [t, y, x] 
    else:
        y = xy_Atom('y', x.pl)
        return [t, y] 
  

def rule_yt(group, word):
    global n_rules
    n_rules += 1
    y, t = word[0], word[1]
    assert t.exp in [1,2]
    if t.exp == 1:
        pl = mat24.pow_ploop(y.pl, 3)
        x, y = xy_Atom('x', pl), xy_Atom('y', pl)
        return [t, y, x] 
    else:
        x = xy_Atom('x', y.pl)
        return [t, x] 


rule_ll = rule_tt




###########################################################################
# Inversions of atoms in the group N
###########################################################################


def iter_inv_p(group, atom):
    global n_inv
    n_inv += 1
    rep = mat24.inv_autpl(atom.rep)
    yield  AutPLoopAtom(atom.tag, rep) 


def iter_inv_xy(group, atom):
    global n_inv
    n_inv += 1
    res = mat24.pow_ploop(atom.pl, 3)
    yield  xy_Atom(atom.tag, res) 


def iter_inv_tl(group, atom):
    global n_inv
    n_inv += 1
    yield tl_Atom(atom.tag, -atom.exp % 3) 


    
###########################################################################
# The group N 
###########################################################################


ERR_ATOM = "Illegal atom %s in constuctor of class MGroupNWord"

def iter_generators_from_atom(atom):
    if atom & -0x80000000 == 0:
        tag = atom >> 28
        data = atom & 0xfffffff
        if tag == 0:    return
        elif tag == 1:  yield AutPLoopAtom('p', data, 0)
        elif tag == 2:  yield AutPLoopAtom('p', 0, data)
        elif tag == 3:  yield xy_Atom('x', data)
        elif tag == 4:  yield xy_Atom('y', data)
        elif tag == 5:  yield tl_Atom('t', data)
        elif tag == 6:  yield tl_Atom('l', data)
        elif data == 0: return
        else:
            raise ValueError(ERR_ATOM % hex(atom))
    else:
        raise ValueError(ERR_ATOM % hex(atom))
            
         
    


parse_functions = {  # deprecated!!!!
            'd' : gen_d,
            'p' : gen_p,
            'x' : gen_xy,         
            'y' : gen_xy,
            'z' : gen_z,
            't' : gen_tl,
            'l' : gen_tl,
}
 
  

class MGroupNWord(AutoGroupWord, AbstractMMGroup):
    def __init__(self,  tag = None, atom = None, *args, **kwds):
        self.seq =  []
        self.reduced = 0
        atoms = iter_mm(self.group, tag, atom)
        for a in atoms:
            for generator in iter_generators_from_atom(a):
                self.seq.append(generator)
            
    def _iter_atoms(self):
        for gen in self.seq:
            yield from gen.as_atoms()

    @property
    def mmdata(self):
        return np.fromiter(self._iter_atoms(), dtype = np.uint32)


class MGroupN(AutoGroup):
    """Models the subgroup N_0 of the Monster MM or one of its covers.

    We may also suport a specific generator l of order 3 of MM \\ N_0 
    so that the monster generated by N_0 and l. The group N_0 or its
    covers are modeled correctly according to their definition in [2].
       
    The present implementation does not model any relations containing
    the generator l. The raison d'etre of generator l is to support 
    operations of MM on its 196884-dimensional represention.
    """
    word_type = MGroupNWord
    is_mmgroup = True
    tags = "pxytl"   
    __instance = None

    def __new__(cls):
        if MGroupN.__instance is None:
             MGroupN.__instance = MGroupNWord.__new__(cls)
        return MGroupN.__instance
 
    rules = {
        r"pp": rule_pp,
        r"xx": rule_xx,
        r"yy": rule_xx,
        r"tt": rule_tt,
        r"ll": rule_ll,
        r"xy": rule_xy,
        r"xp": rule_xp,
        r"yp": rule_yp,
        r"pt": rule_pt,
        r"xt": rule_xt,
        r"yt": rule_yt,
        r"x":  rule_x0,
        r"p":  rule_p0,
        r"t":  rule_tl0,
        r"l":  rule_tl0,
    }

    kernel_rules = {
        r"yx": kernel_rule_yx,
        r"y": kernel_rule_y,
    }
    rules.update(kernel_rules)

    inverter = {
            'p' : iter_inv_p,
            'x' : iter_inv_xy,         
            'y' : iter_inv_xy,
            't' : iter_inv_tl,
            'l' : iter_inv_tl,
    }

    group_name = "MN"
    #kernel = 0  # for compatiblity with older version
    



    def __init__(self):
        """Create an instance of subgroup N_0 of the Monster
		
        Here N_0 is the supgroup the monster defined in 
        [Conw85] We follow the conventions in [Seys19] for
        calculating in N_0. We also define an atom (t,e) in N_0
        which maps to the generator xi**e in [Seys19]; but we
        implement no relations for xi apart from xi**3 = 1.		
		        """
        rules = self.rules 
        super(MGroupN, self).__init__(parse_functions, 
            rules, self.inverter)

    def __call__(self, *args, **kwds):
         return MGroupNWord(*args, **kwds)

    def neutral(self):
         return MGroupNWord()
  



StdMGroupN = MGroupN()
MGroupNWord.group = StdMGroupN
load_group_name(StdMGroupN)
add_to_embedded_classes(MGroupNWord)







from __future__ import absolute_import, division, print_function


####################################################################
# History
####################################################################

VERSION = '0.0.8' # 2022-07-12. Performance improved
# VERSION = '0.0.7' # 2021-12-01. Bugfix in version generation
# VERSION = '0.0.6' # 2021-12-01. Group operation accelerated
# VERSION = '0.0.5' # 2021-08-02. Word shortening in monster implemented
# VERSION = '0.0.4' # 2020-06-15. MSVC compiler is now supported
# VERSION = '0.0.3' # 2020-06-10. bugfixes in code generator
# VERSION = '0.0.2' # 2020-06-04. Order oracle added; bugfixes
# VERSION = '0.0.1' # 2020-05-20. First releae

# Version history must also be updated in the API Reference,
# section 'Version history'.

####################################################################
# Imports
####################################################################


import sys
import os
import re
import subprocess
import numpy as np
from glob import glob
import shutil

import setuptools
from setuptools import setup, find_packages
from collections import defaultdict
from build_ext_steps import Extension, CustomBuildStep, SharedExtension
from build_ext_steps import BuildExtCmd
from build_ext_steps import DistutilsPlatformError, DistutilsSetupError


    

import config
from config import EXTRA_COMPILE_ARGS, EXTRA_LINK_ARGS
from config import ROOT_DIR, SRC_DIR, PACKAGE_DIR, DEV_DIR
from config import REAL_SRC_DIR
from config import C_DIR, PXD_DIR
from config import PRIMES

from codegen_mm_op import mm_op_p_sources
from codegen_mm_reduce import mm_reduce_sources


####################################################################
# Global options
####################################################################


STAGE = 1
PRIMES = PRIMES[:]
# Parse a global option '--stage=i" and set variable ``STAGE``
# to the integer value i if such an option is present.
# Parse a global option "--p=p1,p2,..." to a list PRIMES=[p1,p2,...]
for i, s in enumerate(sys.argv[1:]):
    if s.startswith("--stage="):
        STAGE = int(s[8:])
        sys.argv[i+1] = None
    if s.startswith("--p="):
        PRIMES = list(set(map(int, s[4:].split(","))) | set([3,15]))
        PRIMES.sort()
        sys.argv[i+1] = None
    elif s[:1].isalpha:
        break
while None in sys.argv: 
    sys.argv.remove(None)

#print(STAGE, PRIMES, sys.argv); 

####################################################################
# Delete files
####################################################################


# The following files are before building the extension
# if the command line option -f or --force has been set
ext_delete = [
    os.path.join("C_DIR", "*.*"),
    os.path.join("PXD_DIR", "*.*"),
    os.path.join(PACKAGE_DIR, "*.dll"), 
    os.path.join(PACKAGE_DIR, "*.pyd"),
]


def force_delete():
    """Delete some files before command 'build_ext'"""
    if not "-f" in sys.argv and not  "--force" in sys.argv:
        return
    for file_pattern in ext_delete:
        for file in glob(file_pattern):
            try:
                #print(file)
                os.remove(file)
            except:
                pass
   


####################################################################
# create directories
####################################################################

def make_dir(*args):
    """Create subdirectory if it does not exist

    The path is given by the arguments
    """
    directory = os.path.realpath(os.path.join(*args))
    if not os.path.exists(directory):
        os.makedirs(directory)
    fname = os.path.join(directory, "readme.txt")
    with open(fname, "wt") as f:
        f.write(
"""The files in this directory have been created automatically
or copied from some other place.
So it is safe to delete all files in this directory.
"""
        )   

####################################################################
# extend path
####################################################################

def extend_path():
    sys.path.append(REAL_SRC_DIR)

####################################################################
# Check if we are in a 'readthedocs' environment
####################################################################


on_readthedocs = os.environ.get('READTHEDOCS') == 'True'



codegen_args = ["mockup"] if on_readthedocs else []

####################################################################
# Set path for shared libraries in linux
####################################################################

if not on_readthedocs and os.name == "posix":    
    old_ld_path = os.getenv("LD_LIBRARY_PATH")
    old_ld_path = old_ld_path + ";" if old_ld_path else ""
    new_LD_LIBRARY_PATH = os.path.abspath(PACKAGE_DIR)
    os.environ["LD_LIBRARY_PATH"] =  old_ld_path + new_LD_LIBRARY_PATH 


####################################################################
# Add extensions and shared libraries to package data
####################################################################


if os.name in ["nt"]:
    extension_wildcards =  ["*.pyd", "*.dll"]     
elif os.name in ["posix"]:
    extension_wildcards =  ["*.so"]  
else:   
    extension_wildcards =  []  


package_data = {
        # If any package contains *.txt or *.rst files, include them:
        "mmgroup": extension_wildcards
}




####################################################################
# Desription of the list 'general_presteps'.
#
# This is a list of programs to be run before executing the 'build_ext' 
# command. Each entry of list 'custom_presteps' is a list which we call 
# a program list. A program list ia a list of strings corresponding to 
# a program to be executed with:
#     subprocess.check_call(program_list) . 
# The first entry of a program list is the name of the program to be 
# executed; here sys.executable means the current python version. 
# Subsequents entries correspond to command line arguments.
#
# If the first entry in that list is not a string then it is 
# interpreted as a function to be called with the arguments
# given by the subsequent entries of that list.
####################################################################





pyx_sources = [
    os.path.join(DEV_DIR, "mat24", "mat24fast.pyx"),
    os.path.join(DEV_DIR, "generators", "generators.pyx"),
    os.path.join(DEV_DIR, "mm_basics", "mm_basics.pyx"),
    os.path.join(DEV_DIR, "clifford12", "clifford12.pyx"),
    os.path.join(DEV_DIR, "mm_reduce", "mm_reduce.pyx"),
]

def copy_pyx_sources():
    for filename in pyx_sources:
        shutil.copy(filename, PXD_DIR)

    

general_presteps = CustomBuildStep("Starting code generation",
  [make_dir, "src", "mmgroup", "dev", "c_files"],
  [make_dir, "src", "mmgroup", "dev", "pxd_files"],
  [force_delete],
  [copy_pyx_sources],
  [extend_path],
)
if STAGE > 1:
    general_presteps = CustomBuildStep("Starting code generation",
       [copy_pyx_sources],
       [extend_path],
    )

####################################################################
# We have to divide the code generation process 
# into stages, since a library built in a certain stage may be 
# for generating the code used in a subsequent stage.
####################################################################





####################################################################
# Building the extenstions at stage 1
####################################################################



mat24_presteps = CustomBuildStep("Generating code for extension 'mat24'",
  [sys.executable, "codegen_mat24.py"],
  [sys.executable, "codegen_clifford12.py"],
)

mat24_shared = SharedExtension(
    name = "mmgroup.mmgroup_mat24", 
    sources=[
        os.path.join(C_DIR, "mat24_functions.c"),
        os.path.join(C_DIR, "gen_xi_functions.c"),
        os.path.join(C_DIR, "mm_group_n.c"),
        os.path.join(C_DIR, "gen_leech_type.c"),
        os.path.join(C_DIR, "gen_leech.c"),
        os.path.join(C_DIR, "gen_leech3.c"),
        os.path.join(C_DIR, "gen_leech_reduce.c"),
        os.path.join(C_DIR, "gen_leech_reduce_n.c"),
        os.path.join(C_DIR, "gen_random.c"),
    ],
    libraries = [], 
    include_dirs = [PACKAGE_DIR, C_DIR],
    library_dirs = [PACKAGE_DIR, C_DIR],
    extra_compile_args = EXTRA_COMPILE_ARGS,
    implib_dir = C_DIR,
    define_macros = [ ("MAT24_DLL_EXPORTS", None)],
)


shared_libs_before_stage1 = [
   mat24_shared.lib_name
] if not on_readthedocs else []
# Attribute ``lib_name`` of a instance of clsss ``SharedExtension``
# contains name of the library (or of the import library in Windows) 
# that has to be linked to a program using the shared library.



clifford12_sources = [
    "qstate12", "qmatrix12", "bitmatrix64",  "uint_sort", "xsp2co1", 
    "leech3matrix", "xsp2co1_elem", "involutions",
     "xsp2co1_traces",
]

clifford12_shared = SharedExtension(
    name = "mmgroup.mmgroup_clifford12", 
    sources = [os.path.join(C_DIR, f) + ".c" for f in clifford12_sources],
    include_dirs = [PACKAGE_DIR, C_DIR],
    library_dirs = [PACKAGE_DIR, C_DIR],
    libraries = shared_libs_before_stage1, 
    extra_compile_args = EXTRA_COMPILE_ARGS,
    implib_dir = C_DIR,
    define_macros = [ ("CLIFFORD12_DLL_EXPORTS", None)],
)


shared_libs_stage1 = shared_libs_before_stage1 + [
       clifford12_shared.lib_name
] if not on_readthedocs else []




mat24_extension = Extension("mmgroup.mat24",
        sources=[
            os.path.join(PXD_DIR, "mat24fast.pyx"),
        ],
        #libraries=["m"] # Unix-like specific
        include_dirs = [ C_DIR ],
        library_dirs = [PACKAGE_DIR, C_DIR ],
        libraries = shared_libs_stage1, 
        #runtime_library_dirs = ["."],
        extra_compile_args = EXTRA_COMPILE_ARGS, 
        extra_link_args = EXTRA_LINK_ARGS, 
)

generators_extension = Extension("mmgroup.generators",
        sources=[
            os.path.join(PXD_DIR, "generators.pyx"),
        ],
        #libraries=["m"] # Unix-like specific
        include_dirs = [ C_DIR ],
        library_dirs = [PACKAGE_DIR, C_DIR ],
        libraries = shared_libs_stage1, 
        #runtime_library_dirs = ["."],
        extra_compile_args = EXTRA_COMPILE_ARGS, 
        extra_link_args = EXTRA_LINK_ARGS, 
)


clifford12_extension =  Extension("mmgroup.clifford12",
        sources=[
            os.path.join(PXD_DIR, "clifford12.pyx"),
        ],
        #libraries=["m"] # Unix-like specific
        include_dirs = [ C_DIR ],
        library_dirs = [PACKAGE_DIR, C_DIR ],
        libraries = shared_libs_stage1, 
        #runtime_library_dirs = ["."],
        extra_compile_args = EXTRA_COMPILE_ARGS, 
        extra_link_args = EXTRA_LINK_ARGS, 
)


####################################################################
# Building the extenstions at stage 2
####################################################################



mm_presteps =  CustomBuildStep("Code generation for modules mm and mm_op",
  [sys.executable, "codegen_mm.py"] + codegen_args,
  [sys.executable, "codegen_mm_op.py"] + codegen_args,
)


mm_shared =  SharedExtension(
    name = "mmgroup.mmgroup_mm_basics", 
    sources=[ os.path.join(C_DIR, f) for f in 
        [ "mm_aux.c",  "mm_group_word.c",
          "mm_tables.c","mm_tables_xi.c", "mm_crt.c",
        ]
    ],    
    libraries = shared_libs_stage1, 
    include_dirs = [PACKAGE_DIR, C_DIR],
    library_dirs = [PACKAGE_DIR, C_DIR],
    extra_compile_args = EXTRA_COMPILE_ARGS,
    implib_dir = C_DIR,
    define_macros = [ ("MM_BASICS_DLL_EXPORTS", None)],
)

shared_libs_stage2 = shared_libs_stage1 + [
       mm_shared.lib_name
] if not on_readthedocs else []




mm_extension = Extension("mmgroup.mm",
    sources=[
            os.path.join(PXD_DIR, "mm_basics.pyx"),
    ],
    #libraries=["m"] # Unix-like specific
    include_dirs = [ C_DIR ],
    library_dirs = [ PACKAGE_DIR, C_DIR ],
    libraries = shared_libs_stage2, 
            # for openmp add "libgomp" 
    #runtime_library_dirs = ["."],
    extra_compile_args = EXTRA_COMPILE_ARGS, 
            # for openmp add "-fopenmp" 
    extra_link_args = EXTRA_LINK_ARGS, 
            # for openmp add "-fopenmp" 
)



ext_modules = [
    general_presteps,
    mat24_presteps,
    mat24_shared,
    clifford12_shared, 
    mat24_extension,
    generators_extension,
    clifford12_extension,
    mm_presteps,
    mm_shared, 
    mm_extension, 
]


if STAGE >= 2:
    ext_modules = ext_modules[:1] + ext_modules[-3:]




####################################################################
# Adding the extension for operation modulo p.
#
# Here we build the representation of the monster modulo
# all small primes P in the set PRIMES. 
####################################################################

PYX_SOURCE_P = "mm_op{P}.pyx"


def list_source_files(p):
    """Return list of fully-qualified names of C sources for modulus p

    This list of C source files is required for building the Cython
    extension "mmgroup_mm_op{P}", where {P} is to be replaced by the
    modulus p.

    This function calls function ``mm_op_p_sources(p)`` in module
    ``codegen_mm_op``. Function ``mm_op_p_sources(p)`` returns a list
    of strings of shape "mm{P}_op_word". In case p = 7 this string
    means the C file "mm7_op_word.c".
    """
    sources = []
    for f in mm_op_p_sources(p):
         sources.append(os.path.join(C_DIR, f.format(P = p) + ".c"))
    return sources


mm_op_shared = {}

    
for p in PRIMES:
    mm_op_shared[p] = shared = SharedExtension(
        name = "mmgroup.mmgroup_mm_op%d" % p, 
        sources =  list_source_files(p),
        libraries = shared_libs_stage2, 
        include_dirs = [PACKAGE_DIR, C_DIR],
        library_dirs = [PACKAGE_DIR, C_DIR],
        extra_compile_args = EXTRA_COMPILE_ARGS,
        implib_dir = C_DIR,
        define_macros = [ ("MM_OP%s_DLL_EXPORTS" % p, None)],
    )
    ext_modules.append(shared)

    sources = [os.path.join(PXD_DIR, PYX_SOURCE_P.format(P = p))]
    ext_modules.append(
        Extension("mmgroup.mm%d" % p,
            sources = sources,
            #libraries=["m"] # Unix-like specific
            include_dirs = [ C_DIR ] , 
            library_dirs = [PACKAGE_DIR, C_DIR ],
            libraries = shared_libs_stage2 + [shared.lib_name], 
                # for openmp add "libgomp" 
            #runtime_library_dirs = ["."],
            extra_compile_args = EXTRA_COMPILE_ARGS, 
                # for openmp add "-fopenmp" 
            extra_link_args = EXTRA_LINK_ARGS, 
                # for openmp add "-fopenmp" 
        )
    )


####################################################################
# Building the extenstions at stage 3
####################################################################


if STAGE >= 3:
    ext_modules = ext_modules[:1]



reduce_presteps =  CustomBuildStep("Code generation for modules mm_reduce",
  [sys.executable, "codegen_mm_reduce.py"] + codegen_args,
)


shared_libs_stage2_augmented = shared_libs_stage2 + [
       mm_op_shared[15].lib_name
] if not on_readthedocs else []


mm_reduce =  SharedExtension(
    name = "mmgroup.mmgroup_mm_reduce", 
    sources=[os.path.join(C_DIR, f) + ".c" for f in mm_reduce_sources()],    
    libraries = shared_libs_stage2_augmented, 
    include_dirs = [PACKAGE_DIR, C_DIR],
    library_dirs = [PACKAGE_DIR, C_DIR],
    extra_compile_args = EXTRA_COMPILE_ARGS,
    implib_dir = C_DIR,
    define_macros = [ ("MM_REDUCE_DLL_EXPORTS", None)],
)


shared_libs_stage3 = shared_libs_stage2_augmented + [
       mm_reduce.lib_name
] if not on_readthedocs else []


mm_reduce_extension = Extension("mmgroup.mm_reduce",
    sources=[
            os.path.join(PXD_DIR, "mm_reduce.pyx"),
    ],
    #libraries=["m"] # Unix-like specific
    include_dirs = [ C_DIR ],
    library_dirs = [ PACKAGE_DIR, C_DIR ],
    libraries = shared_libs_stage3, 
            # for openmp add "libgomp" 
    #runtime_library_dirs = ["."],
    extra_compile_args = EXTRA_COMPILE_ARGS, 
            # for openmp add "-fopenmp" 
    extra_link_args = EXTRA_LINK_ARGS, 
            # for openmp add "-fopenmp" 
)




ext_modules += [
    reduce_presteps,
    mm_reduce,
    mm_reduce_extension,
]


####################################################################
# After building the externals we add a tiny little test step.
####################################################################



test_step = CustomBuildStep("import_all",
  [sys.executable, "import_all.py"],
  ["pytest",  "src/mmgroup/", "-v", "-s", "-m", "build"],
)


ext_modules.append(test_step)



####################################################################
# Don't build any externals when building the documentation.
####################################################################


if on_readthedocs:
    ext_modules = [ 
        general_presteps,
        mat24_presteps,
        mm_presteps,
        reduce_presteps,
    ]

   

def read(fname):
    """Return the text in the file with name 'fname'""" 
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


####################################################################
# The main setup program.
####################################################################


### Prelimiary!!!
"""
ext_modules = [
    mat24_presteps,
    clifford12_shared, 
    clifford12_extension,
]
"""

setup(
    name = 'mmgroup',    
    version = VERSION,    
    license='BSD-2-Clause',
    description='Implementation of the sporadic simple monster group.',
    long_description=read('README.rst'),
    author='Martin Seysen',
    author_email='m.seysen@gmx.de',
    url='https://github.com/Martin-Seysen/mmgroup',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 1 - Planning',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        #'Operating System :: Unix',
        #'Operating System :: POSIX',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python',
        #Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        #'Programming Language :: Python :: 3.4',
        #'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        # uncomment if you test on these interpreters:
        # 'Programming Language :: Python :: Implementation :: IronPython',
        # 'Programming Language :: Python :: Implementation :: Jython',
        # 'Programming Language :: Python :: Implementation :: Stackless',
        'Topic :: Scientific/Engineering :: Mathematics',
    ],
    project_urls={
       # 'Changelog': 'yet unknown',
       # 'Issue Tracker': 'yet unknown',
    },
    keywords=[
        'sporadic group', 'monster group', 'finite simple group'
    ],
    python_requires='>=3.6',
    install_requires=[
         'numpy',
    ],
    extras_require={
        # eg:
        #   'rst': ['docutils>=0.11'],
        #   ':python_version=="2.6"': ['argparse'],
    },
    setup_requires=[
        'numpy', 'scipy', 'pytest-runner', 'cython', 'regex',
        # 'sphinx',  'sphinxcontrib-bibtex',
    ],
    tests_require=[
        'pytest', 'scipy', 
    ],
    cmdclass={
        'build_ext': BuildExtCmd,
    },
    ext_modules = ext_modules,
    package_data = package_data,
    include_dirs=[np.get_include()],  # This gets all the required Numpy core files
)


####################################################################
# In linux, use 'auditwheel' afterwards, to build the final wheel.
#
# The current version is still erroneous.
####################################################################



def build_posix_wheel():   
    assert  os.name == "posix" 
    if not on_readthedocs and "bdist_wheel" in sys.argv:
        PROJECT_NAME = r"mmgroup"
        SUFFIX_MATCH = r"[-0-9A-Za-z._]+linux[-0-9A-Za-z._]+\.whl"
        DIST_DIR = "dist"
        w_match = re.compile(PROJECT_NAME + SUFFIX_MATCH)
        wheels = [s for s in os.listdir(DIST_DIR) if w_match.match(s)]
        for wheel in wheels:
            wheel_path = os.path.join(DIST_DIR, wheel)
            args = ["auditwheel", "-v", "repair", wheel_path]
            print(" ".join(args))
            subprocess.check_call(args)



if os.name == "posix":
    ## build_posix_wheel()  # This does not work
    pass





from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals



from random import randint #, shuffle, sample
from functools import reduce
from operator import __or__

import numpy as np
import pytest

from mmgroup.structures.qs_matrix import QStateMatrix, qs_rand_matrix
from mmgroup.structures.qs_matrix import qs_pauli_matrix
from mmgroup.structures.qs_matrix import qs_ctrl_not_matrix
from mmgroup.structures.qs_matrix import qs_phi_matrix
from mmgroup.structures.qs_matrix import qs_ctrl_phi_matrix
from mmgroup.structures.qs_matrix import qs_hadamard_matrix


from mmgroup.tests.test_clifford.test_qs_matrix import compare_complex

from mmgroup.bitfunctions import bitparity

#####################################################################
# Test gate_not
#####################################################################


gate_not_data = [
     (4, [0b000], 0b11),
     (4, [0b110_1001, 0b101_0111, 0b011_0100], 0b11),
]

def gate_not_testdata():
    for cols, data, v in gate_not_data:
        yield QStateMatrix(0, cols, data), v
    for cols in list(range(6)):
        for r in range(cols+3):
            for i in range(3):
                v = randint(0, (1 << cols) - 1)
                yield  qs_rand_matrix(0, cols, r), v

def complex_gate_not(c1, v):
    v &= len(c1) - 1
    c2 = np.copy(c1)
    for i, x in enumerate(c1):
        c2[i ^ v] = x
    return c2
        
@pytest.mark.qstate
def test_gate_not(verbose = 0):
    """Test function ``qstate12_gate_not``. """
    for ntest, (m, v) in enumerate(gate_not_testdata()):        
        if verbose:
            print("TEST gate not %s" % (ntest+1))
            print(m)
            print("m.shape=", m.shape, "v=", hex(v))
        m1 = m.copy().gate_not(v) 
        c, c1 = m.complex().ravel(), m1.complex().ravel()
        c1_ref = complex_gate_not(c, v)   
        if verbose:
            print("output")
            print(m1)
        err = "Method gate_not has failed"
        compare_complex(c1_ref, c1, err)            
        gm = qs_pauli_matrix(m.shape[1], v << m.shape[1]) 
        assert m @ gm == m1, [str(x) for x in (m, gm, m @ gm, m1)]         
            
            
#####################################################################
# Test gate_ctrl_not
#####################################################################


gate_ctrl_not_data = [
     (1, [0b0], 1, 0),
     (4, [0b110_1001, 0b101_0111, 0b011_0100], 0b11, 0b11),
]

def gate_ctrl_not_testdata():
    for cols, data, v1, v2 in gate_ctrl_not_data:
        yield QStateMatrix(0, cols, data), v1, v2
    for cols in list(range(6)):
        for r in range(cols+3):
            for i in range(3):
                vc = randint(0, (1 << cols) - 1)
                v = randint(0, (1 << cols) - 1)
                while bitparity(vc & v):
                    v = randint(0, (1 << cols) - 1)
                m = qs_rand_matrix(0, cols, r)
                yield m, vc, v


def parity32(v):
    v ^= v >> 16
    v ^= v >> 8
    v ^= v >> 4
    v ^= v >> 2
    v ^= v >> 1
    return v & 1

def complex_gate_ctrl_not(c1, vc, v):
    vc &= len(c1) - 1
    v &= len(c1) - 1
    c2 = 0 * c1
    for i, x in enumerate(c1):
        c2[i ^ (v * parity32(vc & i))] += x
    return c2
  
@pytest.mark.qstate
def test_gate_ctrl_not(verbose = 0):
    """Test function ``qstate12_gate_ctrl_not``. """
    for ntest, (m, vc, v) in enumerate(gate_ctrl_not_testdata()):
        if verbose:
            print("TEST gate ctrl_not %s" % (ntest+1))
            print(m)
            print("vc=", hex(vc))
            print("v=", hex(v))
        m1 = m.copy().gate_ctrl_not(vc, v) 
        c, c1 = m.complex().ravel(), m1.complex().ravel()
        c1_ref = complex_gate_ctrl_not(c, vc, v)    
        if verbose:
            print("output")
            print(m1)
        err = "Method gate_ctrl_not has failed"
        compare_complex(c1_ref, c1, err) 
        gm = qs_ctrl_not_matrix(m.shape[1], vc, v)  
        assert m @ gm == m1, [str(x) for x in (m, m @ gm, m1)]         



#####################################################################
# Test gate_phi
#####################################################################



def gate_phi_testdata():
    for cols in list(range(6)):
        for r in range(cols+3):
            for i in range(3):
                v = randint(0, (1 << cols) - 1)
                phi = randint(0,4); 
                m = qs_rand_matrix(0, cols, r)
                yield m, v, phi


phi_mult = [1, 1j, -1, -1j]

def complex_gate_phi(c1, v, phi):
    v &= len(c1) - 1
    f = phi_mult[phi & 3]
    c2 = np.copy(c1)
    for i, x in enumerate(c1):
        c2[i] = x * (f if parity32(i & v) else 1)
    return c2


@pytest.mark.qstate
def test_gate_phi(verbose = 0):
    """Test function ``qstate12_gate_phi``. """
    for ntest, (m, v, phi) in enumerate(gate_phi_testdata()):
        assert 0 <= v < 1 << m.shape[1]
        if verbose:
            print("TEST gate phi %s" % (ntest+1))
            print(m)
            print("v=", hex(v))
            print("phi=", phi & 3)
        m1 = m.copy().gate_phi(v, phi) 
        c, c1 = m.complex().ravel(), m1.complex().ravel()
        c1_ref = complex_gate_phi(c, v, phi)   
        if verbose:
            print("output")
            print(m1)
        err = "Method gate_phi has failed"
        compare_complex(c1_ref, c1, err)            
        gm = qs_phi_matrix(m.shape[1], v, phi)  
        assert m @ gm == m1, [str(x) for x in (m, m @ gm, m1)]
        if (phi & 3 == 2):         
            gm = qs_pauli_matrix(m.shape[1], v)  
            assert m @ gm == m1, [str(x) for x in (m, m @ gm, m1)]         


#####################################################################
# Test gate_ctrl_phi
#####################################################################

gate_ctrl_phi_data = [
     (1, [0b0], 1, 1),
     (2, [0b11], 0b10, 0b11),
]

def gate_ctrl_phi_testdata():
    for cols, data, v1, v2 in gate_ctrl_phi_data:
        yield QStateMatrix(0, cols, data), v1, v2
    for cols in list(range(6)):
        for r in range(cols+3):
            for i in range(3):
                v1 = randint(0, (1 << cols) - 1)
                v2 = randint(0, (1 << cols) - 1)
                m = qs_rand_matrix(0, cols, r)
                yield m, v1, v1



def complex_gate_ctrl_phi(c1, v1, v2):
    v1 &= len(c1) - 1
    v2 &= len(c1) - 1
    c2 = np.copy(c1)
    for i, x in enumerate(c1):
        f = parity32(i & v1) * parity32(i & v2) 
        c2[i] = -x if f else x
    return c2


@pytest.mark.qstate
def test_gate_ctrl_phi(verbose = 0):
    """Test function ``qstate12_gate_ctrl_phi``. """
    for ntest, (m, v1, v2) in enumerate(gate_ctrl_phi_testdata()):
        if verbose:
            print("TEST gate ctrl phi %s" % (ntest+1))
            print(m)
            print("v1=", hex(v1))
            print("v2=", hex(v2))
        m1 = m.copy().gate_ctrl_phi(v1, v2) 
        c, c1 = m.complex().ravel(), m1.complex().ravel()
        c1_ref = complex_gate_ctrl_phi(c, v1, v2)   
        if verbose:
            print("output")
            print(m1)
        err = "Method gate_ctrl_phi has failed"
        compare_complex(c1_ref, c1, err)            
        gm = qs_ctrl_phi_matrix(m.shape[1], v1, v2)  
        assert m @ gm == m1, [str(x) for x in (m, m @ gm, m1)]         



#####################################################################
# Test gate_h
#####################################################################



gate_h_data = [
     (1, [0b0], 1),
     (2, [0b11], 0b11),
]

def gate_h_testdata():
    for cols, data, v in gate_h_data:
        yield QStateMatrix(0, cols, data), v
    for cols in list(range(6)):
        for r in range(cols+3):
            for i in range(3):
                v = randint(0, (1 << cols) - 1)
                m = qs_rand_matrix(0, cols, r)
                yield m, v



def complex_one_gate_h(c1, d):
    c2 = np.copy(c1)
    for i, x in enumerate(c1):
        f = -1 if bool(i & d) else 1
        c2[i] = f * c1[i] + c1[i ^ d]
    return c2


def complex_gate_h(c1, v):
    exp = 0.0
    d = 1
    while (d < len(c1)):
        if v & d:
           c1 = complex_one_gate_h(c1, d)
           exp -= 0.5
        d <<= 1
    return c1 * 2.0**exp


@pytest.mark.qstate
def test_gate_h(verbose = 0):
    """Test function ``qstate12_gate_ctrl_phi``. """
    for ntest, (m, v) in enumerate(gate_h_testdata()):
        if verbose:
            print("TEST gate ctrl phi %s" % (ntest+1))
            print(m)
            print("v=", hex(v))
        m1 = m.copy().gate_h(v) 
        c, c1 = m.complex().ravel(), m1.complex().ravel()
        c1_ref = complex_gate_h(c, v)   
        if verbose:
            print("output")
            print(m1)
        err = "Method gate_ctrl_phi has failed"
        compare_complex(c1_ref, c1, err)            
        gm = qs_hadamard_matrix(m.shape[1], v)  
        assert m @ gm == m1, [str(x) for x in (m, m @ gm, m1)]         







from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals


import pytest

from random import randint

from mmgroup import mat24
mat24fast = mat24


from mmgroup.bitfunctions import bit_mat_mul, v2
from mmgroup.dev.mat24.mat24theta import theta_to_basis_vector


from mmgroup.dev.mat24.mat24_ref import Mat24

#####################################################################
# Test basis of Golay code and cocode against its reference
#####################################################################

@pytest.mark.mat24
def test_basis_is_same():
    for i in range(24):
        assert Mat24.basis[i] == mat24fast.basis[i]
        assert Mat24.recip_basis[i] == mat24fast.recip_basis[i]


@pytest.mark.mat24
@pytest.mark.parametrize("gc", [Mat24, mat24fast])
def test_basis(gc):
    def test_vects():
        for i in range(24): yield gc.basis[i]
        for i in range(80): yield randint(0,0xffffff)
    for v in test_vects():
        ve = gc.vect_to_vintern(v)
        assert ve == bit_mat_mul(v, gc.recip_basis),  map(hex,[i,v,ve])
        ved = gc.vintern_to_vect(ve)
        assert ved == bit_mat_mul(ve, gc.basis),  map(hex,[i,v,ve,ved])
        assert ved == v , map(hex,[i,v,ve,ved])



#####################################################################
# Test function lsbit()
#####################################################################


def lsbit24_testcases():
    for i in range(32):
        i2 = 1 << i
        yield i2
        yield i2-1
        yield i2+1
    yield 0xffffffff
    for i in range(100):
        yield randint(0,0xffffff)


@pytest.mark.mat24
@pytest.mark.parametrize("gc", [Mat24, mat24fast])
def test_lsbit24(gc):
    for x in lsbit24_testcases():
        try:
            v_ref = min(v2(x),24)
        except:
            v_ref = 24
        v = gc.lsbit24(x)              
        assert v == v_ref, (hex(x), v, v_ref)
    print( "test of lsbit determination passed" )


#####################################################################
# Test functions extract_b24() and spread_b24()
#####################################################################


def spread_testcases():
    data = [ (7, 0, 0), (0x311111, 0x101ffe, 0x2e), (0xffffff,)*3 ]
    for d in data:
        yield d
    for k in range(20):
        for i in range(1,4):
            mask = 0xfffffff
            for j in range(i): mask &= randint(0, 0xffffff)
            x = randint(0, 0xffffff)
            l = [(x >> i) & 1 for i in range(24) if (mask >> i) & 1] 
            y = sum(z << i for i, z in enumerate(l))
            yield (mask, x, y)
 

@pytest.mark.mat24
@pytest.mark.parametrize("gc", [Mat24, mat24fast])
def test_spread(gc):
    for mask, x, y in  spread_testcases():
        res = gc.extract_b24(x, mask)
        assert res == y, map(hex, [mask, x, y, res])
        x &= mask
        res = gc.spread_b24(y, mask)
        assert res == x, map(hex, [mask, y, x, res])
    print( "test of 24 bit spread/extract passed" )


#####################################################################
# Test functions related to Golay code conversions
#####################################################################

def gcode_testcases():
    """yields Golay code words as test cases
    """
    yield 0
    for i in range(500):
        yield randint(0, 0xfff)


@pytest.mark.mat24
@pytest.mark.parametrize("gc, ref", [(mat24fast, Mat24)])
def test_gcode(gc, ref):
    for codewordNo in gcode_testcases():
        codeword = gc.gcode_to_vect(codewordNo)
        decoded_codewordNo = gc.vect_to_gcode(codeword) 
        weight =  gc.bw24(codeword)
        assert weight  == 4 * gc.gcode_weight(codewordNo), (
             hex(codeword), gc.gcode_weight(codewordNo) )
        assert decoded_codewordNo == codewordNo
        lg, gcode_bits = gc.vect_to_bit_list(codeword)
        gcode_bits = gcode_bits[:lg]
        assert gcode_bits == gc.gcode_to_bit_list(codewordNo)
        if codewordNo: 
            assert gc.bw24(codeword) in [8,12,16,24], map(hex,[codewordNo,codeword])
        if ref:
            assert gcode_bits == ref.gcode_to_bit_list(codewordNo)
            assert  codeword == ref.gcode_to_vect(codewordNo)
            assert  decoded_codewordNo == ref.vect_to_gcode(codeword)
            assert  weight  == 4 * ref.gcode_weight(codewordNo)
    print("Golay code test passed")


#####################################################################
# Test functions related to Golay code syndrome
#####################################################################


def syndrome_testcases():
    """yields test cases (codewordNo, cocode, t) for Mat24.syndrome

    codewordNo is 12-bit random number indicating a Golay code word
    cocode is a list of 0...4 bit position where to modify the code word
    t is an entry of cocode if len(cocode) = 4 and a random bit position
    (or 24) otherwise.
    """
    testcases = [
          (1, [1], 3), (0x400,[0],0), (0xd06,[0],0)
        # (0x800, [0x800], 3), (0x200,[0],0), (0x683,[0],0)
    ]
    for t in testcases: 
        yield t
    ll = [0,1,1,2,2,2,3,3,3,3,3,4,4,4,4,4]
    for i in range(2000):
        codewordNo = randint(0, 0xfff)
        l = ll[randint(0,len(ll)-1)]
        cocode = []
        while len(cocode) < l:
            i = randint(0,23)
            if not i in cocode:   cocode.append(i)
        t = cocode[0] if len(cocode) == 4 else min(24, randint(0,29))
        yield codewordNo, cocode, t



@pytest.mark.mat24
@pytest.mark.parametrize("gc, ref", [(mat24fast, Mat24)])
def test_syndrome(gc, ref):
    for i in range(24):
        for j in range(i+1,24):
            for k in range(j+1,24):
                v = (1<<i) + (1<<j) + (1<<k)
                if gc.syndrome(v ^ 0xffffff) != v:
                    print( i,j,k, binbv(v), binbv( LexGolayCode.syndrome(v)) )
                    raise ValueError( "decoding error" )

    for codewordNo, cocode, t in syndrome_testcases():
        codeword = gc.gcode_to_vect(codewordNo)
        ## The following stuff is already done in test_gcode()
        #decoded_codewordNo = gc.vect_to_gcode(codeword) 
        #weight =  gc.bw24(codeword)
        #assert weight  == 4 * gc.gcode_weight(codewordNo), (
        #     hex(codeword), gc.gcode_weight(codewordNo) )
        #assert decoded_codewordNo == codewordNo
        
        if codewordNo: 
            assert gc.bw24(codeword) in [8,12,16,24], map(hex,[codewordNo,codeword])
        synd = sum([1 << i for i in cocode])
        vector = codeword ^ synd
        encoded = gc.vect_to_vintern(vector)
        decoded = gc.vintern_to_vect(encoded)
        assert vector == decoded

        syndrome = gc.syndrome(vector,t)
        assert syndrome  == synd , map(hex, [codeword, synd,t, syndrome])
        coc = gc.vect_to_cocode(vector)
        assert coc == gc.vect_to_cocode(synd)    
        sample_v =  gc.cocode_to_vect(gc.vect_to_cocode(vector))
        assert coc == gc.vect_to_cocode(sample_v) 
        assert gc.cocode_to_bit_list(coc, t) == sorted(cocode)
        if ref:
            assert  ref.cocode_to_bit_list(coc, t) == sorted(cocode)
            assert  syndrome == ref.syndrome(vector,t), (hex(syndrome), t) 
            assert  coc == ref.vect_to_cocode(vector)
            ## The following stuff is already done in test_gcode()
            #assert  codeword == ref.gcode_to_vect(codewordNo)
            #assert  decoded_codewordNo == ref.vect_to_gcode(codeword)
            #assert  weight  == 4 * ref.gcode_weight(codewordNo)
        if len(cocode) == 4:
            with pytest.raises(ValueError):
                gc.syndrome(vector,24)
            t = gc.cocode_to_sextet(coc)
            assert t == ref.cocode_to_sextet(coc)
            assert sum([1 << i for i in t]) == 0xffffff
            for i in range(0, 24, 4):
                v1 = sum([1 << j for j in t[i:i+4]])
                assert gc.vect_to_cocode(v1) == coc
    print( "Golay code syndome decoding test passed" )





#####################################################################
# Test functions related to Golay cocode weight
#####################################################################


def cocode_weight_testcases(gc):
   yield 0
   for i in range(24):
       yield gc.vect_to_cocode(1 << i)
   for i in range(500):
       yield randint(0, 0xfff)



@pytest.mark.mat24
@pytest.mark.parametrize("gc, ref", [(mat24fast, Mat24)])
def test_cocode_weight(gc, ref):
    for c1 in cocode_weight_testcases(gc):
        w = gc.bw24(gc.cocode_syndrome(c1, 0))
        assert ref.cocode_weight(c1) ==  w
        assert gc.cocode_weight(c1) ==  w


#####################################################################
# Test functions related to octad numbering
#####################################################################


def octad_testcases():
    testcases = [
        1,  0x800, 0x903
    ]
    for c in testcases: yield c
    for i in range(1000):
        yield randint(0,0xfff)   


@pytest.mark.mat24
@pytest.mark.parametrize("gc, ref", [(mat24fast, Mat24)])
def test_octads(gc, ref):    
    for c in octad_testcases():
        try:
            oct =  gc.gcode_to_octad(c)
        except:
            oct = None
        v = gc.gcode_to_vect(c)
        if oct is None:
            assert gc.bw24(v) in [0,12,16,24], (map(hex,[c, v]))
            continue
        assert 0 <= oct < 759, (hex(c), hex(v), oct)
        assert gc.bw24(v) == 8
        assert gc.vect_to_octad(v) == oct
        assert gc.gcode_to_octad(c) == oct
        assert gc.octad_to_gcode(oct) == c, (oct, hex(c))
        assert gc.octad_to_vect(oct) == v, (hex(c), oct, hex(v))
        ll, bitlist = gc.vect_to_bit_list(v)
        bitlist = bitlist[:ll]
        assert v == sum([1 << i for i in bitlist])
        if ref:
            assert  oct == ref.gcode_to_octad(c)
            assert ref.vect_to_octad(v) == oct
            assert ref.gcode_to_octad(c) == oct
            assert ref.octad_to_vect(oct) == v, (hex(c), oct, hex(v))
            assert ref.octad_to_gcode(oct) == c, (oct, hex(c))

    print( "Golay code octad test passed" )






#####################################################################
# Test functions related to suboctads 
#####################################################################


def suboctad_testcases(gc):
    """yield testcases (octad, suboctad, status) for suboctad functions

    octad is an octad, suboctad is a suboctad, both in vector
    representation. status is 0 if test should not fail.
    status = 2  if octad is not a good octad
    status = 1  is suboctad is not a suboctad of the octad
    """
    testdata = [
      #   (0x00ffff, 0x110000, 0),
      #   (0x00ffff, 0x001111, 1),
      #   (0xeed2dd, 0x481001, 1),
    ]
    for d in testdata:
        yield d
    for i in range(10000):
        oct = gc.gcode_to_vect(randint(1,0xfff)) 
        short_oct = oct if gc.bw24(oct) in [8, 24] else oct ^ 0xffffff
        cocodev = randint(0, 0xffffff)
        sub = gc.cocode_syndrome(cocodev,  gc.lsbit24(short_oct))
        if gc.bw24(short_oct) == 8: 
            status = sub & short_oct != sub or gc.bw24(sub) & 1
        else:
            status = 2
        yield  oct, sub, status
    for i in range(200):
        oct = gc.octad_to_vect(randint(0,758)) 
        sub = randint(0, 0xffffff) & oct
        status = gc.bw24(sub) & 1
        while (i > 50 and status):
            sub = randint(0, 0xffffff) & oct
            status = gc.bw24(sub) & 1
        sign = randint(-1, 0)
        yield (oct ^ sign) & 0xffffff, sub, status
                


@pytest.mark.mat24
@pytest.mark.parametrize("gc, ref", [(mat24fast, Mat24)])
def test_suboctads(gc, ref):
    for octad, suboctad, status in suboctad_testcases(gc):
        #print("Testing suboctad", hex(octad), hex(suboctad), status)
        v = gc.vect_to_gcode(octad)
        c = gc.vect_to_cocode(suboctad)
        last_u_sub = 0
        if status == 0:
            u_sub = gc.cocode_to_suboctad(c,v)
            c1 = gc.suboctad_to_cocode(u_sub,v)
            syn1 = gc.cocode_syndrome(c1, gc.lsbit24(octad))
            assert c == c1, map(hex,[octad, suboctad, u_sub, syn1, status])
            w = gc.suboctad_weight(u_sub)
            assert w == (gc.bw24(suboctad) >> 1) & 1
            scal_ref = (gc.suboctad_weight(u_sub ^ last_u_sub)
              ^ gc.suboctad_weight(u_sub) ^ gc.suboctad_weight(last_u_sub))
            assert gc.suboctad_scalar_prod(u_sub, last_u_sub) == scal_ref
            if ref:
                assert u_sub == ref.cocode_to_suboctad(c,v)
                assert c == ref.suboctad_to_cocode(u_sub,v)
                assert w == ref.suboctad_weight(u_sub)
                assert ref.suboctad_scalar_prod(u_sub, last_u_sub) == scal_ref
            last_sub = u_sub
        else:
            try:
                gc.cocode_to_suboctad(c,v)
            except:
                #print("gc.cocode_to_suboctad failed as expected")
                pass
            else:
                syn1 = gc.syndrome(suboctad, gc.lsbit24(octad))
                print ("Bad suboctad", list(map(hex,[octad, suboctad, syn1, status])))
                raise ValueError("cocode_to_suboctad() should fail bud didn't.")
            if status == 2:
                try:
                    gc.suboctad_to_cocode(randint(0,63),v)
                except:
                    pass
                else:
                    print ("Bad octad", list(map(hex,[octad, suboctad, v, c, status])))
                    raise ValueError("suboctad_to_cocode() should fail bud didn't.")
    print("Golay code suboctad test passed")




